#!/usr/bin/env python
#
# This is a regen tool to go from APT query to all files that can
# come of it:  .pweb, .peng, .jeng
#
# Assume:  the pyetc test repo contains the latest correct APT query sets
#

import os, re, sys, shutil
from pandeia.engine.helpers.bit import bit
from pandeia.engine.helpers.bit.pyetc_util import write_dict, fast_write_dict
from pandeia.engine.helpers.peng import peng3jeng


def get_all_queries(repo_dir_to_apt_queries):
    retval = {}
    for inst in ('acs', 'cos', 'stis', 'wfc3'):
        queryfname = repo_dir_to_apt_queries+'/'+inst+'.txt'
        assert os.path.exists(queryfname), 'Could not find: '+queryfname
        with open(queryfname, 'r') as f:
            iqueries = [q.strip() for q in f.readlines() if len(q.strip()) > 0]
            retval[inst] = iqueries
    return retval


any_unsafe = re.compile(r'[^a-zA-Z0-9_=\.\-]')
def filenamify(qstr):
    """
    Convert some chars to things that work well in file/dirnames, but do so
    in a way that can be fully back-convertable in case we want to convert back.
    Note if the conversion back is done, it muct be done in the reverse order!
    """

    fixed = qstr

    assert '_d_' not in fixed, 'Cannot use "_d_" as a replacement for "." in: '+qstr
    fixed = fixed.replace('.','_d_')

    assert '.' not in fixed, 'Cannot use "." as a replacement for "&" in: '+qstr
    fixed = fixed.replace('&','.')

    assert '_s_' not in fixed, 'Cannot use "_s_" as a replacement for "/" in: '+qstr
    fixed = fixed.replace('/','_s_')

    # make sure the rest is just alpha and numeric and dash/equal/under/dot
    assert any_unsafe.search(fixed) == None, 'Need to handle this query: '+qstr
    return fixed


def write_out_files_for_queries(queries_dict, out_dir, overwrite=True):
    """ Fo each query, dump all the files """
    # logic to show output but reduce debugging
    last_shown = ''
    # go by instrument
    insts = sorted(queries_dict.keys())
    for inst in insts:
        # then by each query
        for query in queries_dict[inst]:
            # output
            line = f'{inst}: {query}'
            line = line[0:line.find('requestCount')-1]+' ...'
            if line != last_shown:
                print(line)
                last_shown = line
            # dirs and fnames
            basedir = out_dir+'/'+inst
            if not os.path.isdir(basedir):
                os.mkdir(basedir)
            basenm = basedir+'/'+filenamify(query)
            out_pweb = basenm+'.pweb'
            out_peng = basenm+'.peng'
            out_jeng = basenm+'.jeng'
            for fname in (out_pweb, out_peng, out_jeng):
                if os.path.exists(fname):
                    if overwrite:
                        os.remove(fname)
                    else:
                        raise IOError('File exists: '+fname)
            # pweb & peng: do the conversions, write em
            pweb_dict, peng_dict = bit.query_to_peng(query)
            fast_write_dict(pweb_dict, out_pweb, comment=f'apt query: {query}')
            fast_write_dict(peng_dict, out_peng, comment=f'apt query: {query}')
            # jeng: convert and write
            if inst == 'stis' and 'centralWavelength' not in query: # some dont work yet
                jeng_dict = peng3jeng.pyetc_in_dict_to_pandeia_in_dict(peng_dict, None, pweb_dict=pweb_dict)
                fast_write_dict(jeng_dict, out_jeng, comment=f'apt query: {query}')


def main():

    pandeia_refdata = os.environ["pandeia_refdata"]
    repo_dir_to_apt_queries = f"{pandeia_refdata}/../pyetc/pyetc/etc_web/bit/apt_queries"
    out_dir = f"{pandeia_refdata}/../bit_output"
    os.makedirs(out_dir, exist_ok=True)
    shutil.rmtree(out_dir) # in case it exists but is full
    os.makedirs(out_dir, exist_ok=True)

    qdict = get_all_queries(repo_dir_to_apt_queries)
    write_out_files_for_queries(qdict, out_dir)

    print('Done!')


#
# main routine
#
if __name__=='__main__': # in case something else imports this file

    main()



'''
This code was copied from the straylight area of the BMG svn repo

These globals were refactored from the top of the straylight accessor because
some of them are also used by the thermal accessor.
'''


import os
from pandeia.engine.helpers.background.jwst.mod_healpix_func import *

bmg_url = os.environ.get('BMG_URL', 'NONE')

D2R = math.pi / 180.
R2D = 180. / math.pi

# C structure used to write the stray light binary file.
# One record per file.
#    typedef struct {
#              double pix_ra;
#              double pix_dec;
#              double pos[3];
#              double nonzodi_bg[SL_NWAVE];  // The in-field ism+cib for each pixel.
#            } sl_pix;
#              Then for every day with the target in the FOR,the following three items are written.
#              double day;
#              double zodi_bg[SL_NWAVE];  //  The combined in-field zodi for each pixel.
#              double stray_light_bg[SL_NWAVE];  //The equivalent in-field background from stray light from zodi, stellar,ism, cib

# Number of days in a year
NUM_DAYS = 366

'''
In principle, this wavelist should be identical (ideally, read from) the standard
wavelength set delivered with the BMG in bgmodel/inputs/std_spectrum_wavelengths.txt.

In practice, it must be identical to the waveset used BACK WHEN the cached data was
generated by other code in the straylight area of the BMG svn repo.

Note that the comment above, which describes the structure of that file, refers to [SL_NWAVE].
'''

#-----------------------------------------------------------------------------------
# THE WAVELENGTH ARRAY:
#
# Copied directly from pandeia/backgrounds/straylight/std_wavelengths.c
#
# NOTE!!! when this is changed it affects THREE (3) hard-coded places:
# 1) here
# 2) pandeia/backgrounds/straylight/std_wavelengths.c (the wave array)
# 3) pandeia/backgrounds/bmg/src/libBG/bmg_user.h (the value of SPECTRUM_ELEMENTS)
#
# (also edit all counterparts in pandeia/backgrounds/bmg_for_straylight)
#-----------------------------------------------------------------------------------

wavelist = [
    0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1,
    2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.7, 4.0, 4.25,
    4.5, 4.75, 5.0, 5.1, 5.2, 5.3, 5.4, 5.5, 5.7, 5.9, 6.1, 6.3, 6.5, 6.7, 6.9, 7.1, 7.3,
    7.5, 7.7, 7.9, 8.1, 8.3, 8.5, 8.7, 8.9, 9.1, 9.3, 9.5, 9.7, 9.9, 10.1, 10.3, 10.5,
    10.7, 10.9, 11.1, 11.3, 11.5, 11.7, 11.9, 12.1, 12.3, 12.5, 12.7, 12.9, 13.1, 13.3,
    13.5, 13.7, 13.9, 14.1, 14.3, 14.5, 14.7, 14.9, 15.1, 15.3, 15.5, 16.5, 17.5, 18.5,
    19.5, 20.5, 21.5, 22.5, 23.5, 24.5, 25.5, 26.5, 27.5, 28.5, 29.5, 30.5, 31.0
]

SL_NWAVE = len(wavelist)
IDX_OF_3P5 = wavelist.index(3.5)  # will throw an exception if fails (which we want)

nside = 128
Npix = 12 * nside * nside



import os

import pandeia.engine.helpers.bit.pyetc_util as util
util.INST_DATADIR = os.path.dirname(__file__)

from pandeia.engine.helpers.bit.instruments.telescopes import TELESCOPES
#####

def import_instrument_module( instrument, module='' ) :
    '''
    import 'instruments.hst.%s.%s%s'%(name,name,which)

    for example:
       engine_module = import_instrument_module( 'acs', '' )
       web_module    = import_instrument_module( 'acs', '_web' )

    '''
    # instrument is the name of the instrument that we are interested in.
    # After #1170, web code lives under a 'web' subpackage. Note also that
    # after the engine refactoring, this method is only called with a
    # module='_web' option. Unit tests call it with module='', and nobody
    # else uses the option.
    if module == '_web':
        name = 'pandeia.engine.helpers.bit.instruments.hst.%s.web.%s%s' % ( instrument, instrument, module )
    else:
        name = 'pandeia.engine.helpers.bit.instruments.hst.%s%s' % ( instrument, module )

#   try :
    return util.dynamic_import(name) # for BIT runs we can handle if this crashes here to highlight any problem
#   except ImportError:
#       pass

    # We could try other possible names here in a similar manner.  When users
    # can provide their own instrument package, it will be convenient to
    # allow them to put the package someplace other than pyetc.instruments.hst
    #

    # if we get to the end, re-raise the last import error
    raise


#####

# This is a short-lived cache for get_instrument_data; it allows us to
# get_instrument_data() whenever we want the data, but we don't have to
# actually read the file multiple times.
instrument_data_cache = { }

def get_instrument_data( instrument, name ) :
    '''
    Read an instrument-specific data file.
    (replaces most uses of etc_engine.util.read_dict())

    For example:
        d = instruments.get_instrument_data( 'acs', 'sharpness' )

    This capability is also present under the name get_data() in each
    instrument-specific web module:

        m = instruments.import_instrument_module( 'acs', '_web' )
        d = m.get_data( 'sharpness' )

    In either case, the file is only read once; the same copy of the
    file data is returned each time.  (Take care not to alter the
    returned object.)

    '''
    index = ( instrument, name )

    if index in instrument_data_cache :
        return instrument_data_cache[index]

    # Find the data directory for this instrument
    dirname = os.path.join(util.INST_DATADIR, 'hst', instrument)

    # read the data
    fname = '%s/%s_%s.dat' % ( dirname, instrument, name )

    datadict = util.read_dict(fname)

    # cache it and return it
    instrument_data_cache[index] = datadict

    return datadict

def get_alt_instrument_data( instrument, name ) :
    '''
    Read an instrument-specific data file.
    (replaces most uses of etc_engine.util.read_dict())

    For example:
        d = instruments.get_instrument_data( 'acs', 'sharpness' )

    This capability is also present under the name get_data() in each
    instrument-specific web module:

        m = instruments.import_instrument_module( 'acs', '_web' )
        d = m.get_data( 'sharpness' )

    In either case, the file is only read once; the same copy of the
    file data is returned each time.  (Take care not to alter the
    returned object.)

    This function should be called when there is a possibility that
    an alternate version of the same data file exists in the
    telescope/instruments/ sub-directory. For now we must code this
    as two separate functions since this module is used by both the
    old and the new versions of the engine code.

    '''
    index = ( instrument, name )

    if index in instrument_data_cache :
        return instrument_data_cache[index]

    # Find the data directory for this instrument
    dirname = os.path.join(util.INST_DATADIR, instrument)

    # read the data
    fname = '%s/%s_%s.dat' % ( dirname, instrument, name )

    datadict = util.read_dict(fname)

    # There might be a new version of he file in pyetc/instruments/.
    # Read it and override former version just read.
    new_name = fname.replace('instruments/hst/%s'%instrument,
                             'instruments/hst/%s/%s/'%(TELESCOPES[instrument],instrument))
    try:
        datadict = util.read_dict(new_name)
    except IOError as e:
        pass

    # cache it and return it
    instrument_data_cache[index] = datadict

    return datadict

#####

# Wrapper to the specific
# <instr>_obsmode.make_<instr>_<sciencemode>_obsmode function

def make_obsmode_test_interface(form_input):
    """This is a test rig to find a back door into instrument specific
    obsmode functions.  For non-test code, you would call the correct
    function directly from the relevant instrument-specific code.
    """

    instr = form_input['instrument'].lower()
    mode = form_input['science_mode'].lower()

    m = import_instrument_module(instr, '_web')

    funcname = 'make_' + instr + '_' + mode + '_obsmode'
    if funcname in m.__dict__ :
        fn = m.__dict__[funcname]
        ans = fn(form_input)
        return ans

    raise NotImplementedError('Instrument/mode combination %s, %s not supported'%(instr,mode))

#####

#Wrapper to the specific <instr>.instrument_from_form function
def instrument_from_form(instr_name, science_mode, form_input):
    """Call the instrument-specific function to process the web form inputs
    into the instrument-specific dictionary.

    instr_name:   lower case string
    science_mode: lower case string
    form_input:   dictionary containing form inputs in proper type
    """

    #This is specific to each instrument_sciencemode combination
    instr_namel = instr_name.lower()

    m = import_instrument_module(instr_namel, '_web')
    idict = m.instrument_from_form(science_mode, form_input)
    return idict, idict.pop("wodict", {})

#####

# This list of all instruments is here for testing only.
test_mode_all_instruments = (
    'acs',
    'cos',
    'stis',
    'wfc3ir',
    'wfc3uvis',
)



short_name = 'wfc3ir'

long_name = 'Wide Field Camera 3 - Infrared'

index_page_text = "Imaging and spectroscopy ETCs are currently offered for the IR channels. "

modes = [
    ( 'imaging',            __name__ + '.web.imaging' ),
    ( 'spectroscopic',      __name__ + '.web.spectroscopic' ),
]


import pandeia.engine.helpers.bit.instruments.hst.default as default
mode_names = default.default_mode_names

short_mode_names = default.default_short_mode_names

import pandeia.engine.helpers.bit.instruments as pi
import pandeia.engine.helpers.bit.instruments.data

location=__file__

data = pi.data.Data('hst', 'wfc3ir', location=location)





#TODO web cleanup:
# there is some duplication in between WFC3 UVIS and IR modules.

import pandeia.engine.helpers.bit.pyetc_util as util
import pandeia.engine.helpers.bit.instruments

#
# read a dictionary from one of the .dat files
def get_data( name ) :
    return pandeia.engine.helpers.bit.instruments.get_instrument_data( 'wfc3ir', name )

FILTERS = get_data("config")["filters"]

FILTER_MAP = {
    'narrow': 'wfc3_filter_n',
    'medium': 'wfc3_filter_m',
    'wide':   'wfc3_filter_w',
    'quad':   'wfc3_filter_q',
}


def instrument_from_form(science_mode, form_input):
    """Function to extract values of interest from the web form
    data and return a smaller dictionary which will be used to
    update the engine_input dictionary.
    """

    idict = dict( nreads = form_input['crsplit'],
                  detector='ir'
                  )
    #check that crsplit is valid
    if idict['nreads'] < 1:
        raise ValueError("CRSPLIT must be greater than 0")

    #Web output dictionary that will be stuffed into db
    idict['wodict'] = wodict = util.stable_dict()
    #the wfc3 IR detectors are ccd-like
    wodict['detector_type'] = 'ccd'
    wodict['IR'] = True
    wodict['HeliumStandard'] = form_input['HeliumStandard']

    if science_mode == 'imaging' or science_mode == 'scimaging':
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'],myfilter =make_wfc3ir_imaging_obsmode(form_input)
        wodict['filter_info']=FILTERS[myfilter]
        wodict['filter_name']=myfilter
        idict['filter'] = myfilter

        return idict

    if science_mode == 'spectroscopic' or science_mode == 'scspectroscopic':
        #convert the obswave in the form from microns to angstroms
        idict['obswave']=form_input['obswave'] * 10000.
        idict['grating']=form_input['disperser'].lower()[-4:]

       #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'] = make_wfc3ir_spectroscopic_obsmode(form_input)

        return idict
    else:
        raise NotImplementedError('%s mode not implemented for wfc3ir'%science_mode)
    return idict


def make_wfc3ir_imaging_obsmode(form_input):
    #Based on the detector, we decide which filters to pick out.
    myfilter = form_input['irfilt0'].lower()
    obsmode='wfc3,ir,'+myfilter
    return  obsmode, myfilter

def make_wfc3ir_spectroscopic_obsmode(form_input):
    myfilter = form_input['disperser'].lower()[4:]

    #this mode ONLY has grisms, so assume filter is a grism to start with
    obsmode='wfc3,ir,'  + myfilter
    return obsmode



import errno, re, sys

from pandeia.engine.helpers.bit.pyetc_util import read_dict, PYSYN_CDBS
from pandeia.engine.helpers.bit import instruments
from pandeia.engine.helpers.bit import string_constants as sc
from pandeia.engine.helpers.bit.extraction import extraction_info
from pandeia.engine.helpers.bit.sky import Sky
from pandeia.engine.helpers.bit.source import Source

# hard-coding subsystem since this code is ONLY here in pandeia to run the BIT
#subsystem = 'etc'
subsystem = 'bit'

# We need to know what telescope to use to find the MJD for the current
# calculation.  This version of the ETC does not really understand multiple
# telescopes, so we assume that we are only configured with one.
telescope = 'hst'

# This is the MJD dict, lazily loaded from the $CDBS/etc/mjd.dat
mjd_dict = None

# obtain the MJD to use for a particular configuration
#
# subsystem normally comes from etc_web.config.subsystem
#
# telescope is which telescope to use.  Since etc_web doesn't currently
# understand multiple telescopes, use the global value telescope from
# this file.
#
# instrument is the name of the instrument.
#
def _handle_MJD( subsystem, telescope, instrument, engine_input ):

    # lazy load mjd_dict
    global mjd_dict
    if mjd_dict is None :
        try :
            mjd_dict = read_dict( PYSYN_CDBS + '/etc/mjd.dat' )
        except IOError as e :
            if e.errno != errno.ENOENT :
                raise
            # no file means there are no entries
            mjd_dict = { }

    # The priorities are:
    # 1. a fully specified entry
    # 2. a subsystem/telescope entry
    # 3. None
    try :
        mjd = mjd_dict[ ( subsystem, telescope, instrument ) ]
    except KeyError :
        try :
            mjd = mjd_dict[ ( subsystem, telescope ) ]
        except KeyError :
            return

    # try :
    #     mjd = mjd_dict[ ( subsystem, telescope, instrument ) ]
    # except KeyError, e :
    #     raise exceptions.DataError("Invalid entry for MJD calculation. Exception is: " + str(e))

    # add MJD to obsmode
    engine_input['obsmode'] += ',MJD#' + str(mjd)


#TODO web cleanup:
#  this function is too long; does too many things.

def process_form_input(instrument, science_mode, form_input, request_dir):
    """ Process the form input dictionary to produce the inputs that the
    engine wants. Return the engine input."""

    # web_output is computation performed here in the web code
    web_output = {}

    # engine input is input to be passed on to the engine
    engine_input = dict()

    ## bug: Apparently, these should be in all of the input forms, but are not.
    #TODO web cleanup:
    #  so, fix it! make sure all forms have 'gyromode', and remove this statement.
    if not 'gyromode' in form_input:
        form_input['gyromode'] = 'normal'

    #and the sky expression & descriptions
    sky = Sky(formdata=form_input)
    sky.make_synexpr()
    engine_input['sky_expr'] = sky.expr
    engine_input.update(sky.descr)

    # instrument_from_form() calls into the instrument
    # specific instrument_from_form(), which converts selected fields
    # from the form input into a dictionary of values
    # to pass on to the engine.
    #
    # obsmode is one of those values.
    idict, wodict = instruments.instrument_from_form(instrument,
                                                     science_mode,
                                                     form_input)
    engine_input.update(idict)
    web_output.update(wodict)

    # add MJD to the instrument obsmode
    _handle_MJD( subsystem, telescope, instrument, engine_input )

    # Create the source expression. This must be created after the
    # instrument-related values have been populated in order to support
    # normalization in the observation bandpass.

    # Due to a bug in the pysynphot parser, the obsmode has to be
    # validated against problematic aperture specifications. Do that first.
    normalization_obsmode = _validate_obsmode(engine_input.get('aperture'),
                                              engine_input.get('obsmode'),
                                              form_input.get('fftype')
                                              )

    src = Source(request_dir=request_dir, formdata=form_input,
                 obsmode=normalization_obsmode)

    # Use this to populate the relevant engine_input and web_output items.
    engine_input['source_expr'] = src.expr
    engine_input['target_lines'] = src.lines
    web_output['source_help'] = src.help

    #Create the extraction info

    scan_mode = science_mode in [sc.SCIMAGING, sc.SCSPECTROSCOPIC]

    # The extraction module looks for the detector in form_input
    #TODO web cleanup:
    # put it in all the forms, then remove this hack
    form_input['detector'] = engine_input['detector']
    engine_input['extraction'] = extraction_info(formdata=form_input, scan_mode=scan_mode)

    #The extraction info wants the detector: add it.
    #TODO web cleanup:
    # couldn't this be done by extraction_info itself ?
    try:
        engine_input['extraction']['detector'] = engine_input['detector']
    except KeyError:
        pass

    #Treat some specific elements in special ways
    engine_input['instrument'] = instrument
    engine_input['science_mode'] = science_mode

    if "Time" in form_input:
        # Normal operation.
        engine_input['time'] = form_input['Time']

    if not scan_mode:
        #If the keys are spelled the same, we can just copy them
        klist = ['SNR',
                 'simmode']
        for key in klist:
            engine_input[key] = form_input[key]

        if 'spectroscopic' in science_mode:
            web_output['spectroscopic'] = True
        else:
            web_output['imaging'] = True

        web_output['mode'] = science_mode

    else:
        _handle_scan_mode(engine_input, form_input, web_output)

    # Rampfilter needs to be set to imaging;
    # otherwise use the science_mode as is.
    if science_mode == 'rampfilter':
        web_output['mode'] = 'imaging'

    return engine_input, web_output


def _handle_scan_mode(engine_input, form_input, web_output):
    """ Handle the specifics associated with scan mode.

    Scan mode supports two possible user inputs:
      - user enters scan rate and scan length. The code in here
        computes Time, which is then fed to the ETC engine to
        compute SNR.
      - user enters Time and scan length. Time is fed to the ETC
        engine as above to compute SNR. The code in here computes
        the scan rate for reporting and warning purposes only.

    The input dictionaries are modified in place.

    Parameters
    ----------
    engine_input: dict
      dict with the engine input
    form_input: dict
      dict with the form input associated with the web request
    web_output: dict
      dict with values used by the output templates

    """
    # Get scan rate from the input request. This might be
    # overwritten below, depending on other input settings.
    engine_input['scrate'] = form_input['scRate']

    # There are two possible sources of scan length in the input
    # page. Pick the one selected by the 'simmode' radio button.
    key = 'scLengthFromRate'
    if form_input['simmode'] == 'SNR':
        key = 'scLengthFromTime'
    engine_input['sclength'] = form_input[key]

    # Time is computed once and for all from the scan length and rate.
    # Or, it's just read from the Time field.
    # Page and/or validation code should handle limiting cases.
    simmode = form_input['simmode']
    if simmode == 'SNR_from_rate':
        engine_input['time'] = float(engine_input['sclength']) / float(engine_input['scrate'])
    else:
        engine_input['time'] = form_input['Time']
        engine_input['scrate'] = engine_input['sclength'] / engine_input['time']

    # Time is reported directly. No inverse computation is supported.
    web_output['time'] = engine_input['time']

    # SNR is passed to the engine input but never used in scan mode.
    # This only fulfills the engine API requirements.
    engine_input['SNR'] = form_input['SNR']

    # We only support the direct SNR calculation.
    engine_input['simmode'] = 'SNR'

    # report correct mode at output.
    if 'scspectroscopic' in engine_input['science_mode']:
        web_output['scspectroscopic'] = True
    else:
        web_output['scimaging'] = True

    # Some science modes look better when reported
    # in a more human-readable form.
    if engine_input['science_mode'] == 'scimaging':
        web_output['mode'] = 'imaging scan'
    elif engine_input['science_mode'] == 'scspectroscopic':
        web_output['mode'] = 'spectroscopic scan'


def _validate_obsmode(aperture, obsmode, normtype):
    """ Validate for avoidance of pysynphot parser bug (pysyn #238)

    Some apertures/slits are spelled with digits followed by letters,
    and this causes the pysynphot parser to incorrectly break up the
    keyword which then causes an exception either in the parsing phase
    or in the graph table lookup. This function removes problematic
    apertures from the obsmode string.

    Parameters
    ----------
    aperture: string
    obsmode: string
    normtype: string

    Returns
    -------
    result: string

    """
    result = obsmode
    # The bug appears only when normalizing by bandpass.
    if normtype != 'fNormalizeByBandpass':
        return result

    # Not all obsmodes have apertures.
    if aperture is None:
        return result
    if aperture not in obsmode:
        return result

    # Not all apertures are problems.
    # Use re.match to determine whether it starts with a digit.
    problem = re.match("[0-9]", aperture)
    if problem:
        raise ValueError("Normalization in observation bandpass that includes aperture %s is not supported at this time."%aperture)

    return result



import datetime
import os

from pandeia.engine.helpers.background.hst import util
from pandeia.engine.helpers.background.hst import sky_angles as angles
from pandeia.engine.custom_exceptions import BackgroundError, RangeError, UserInputError

from pandeia.engine.utils import recursive_subclasses, merge_data

'''
keep the synphot expression generation section of the code in sync with 
https://github.com/spacetelescope/pyetc/blob/master/pyetc/etc_web/etc/sky.py
'''
#TODO: sky.dat needs to migrate to CDBS & have a versioned name
MODELS = util.read_dict(os.path.join(os.path.dirname(__file__),
                                'sky.dat'))

class SkyElement(object):
    """Base class for actual elements of the sky model"""
    #Subclasses will override this value
    formkeys = []


    def __init__(self, formdata=None):
        """formdata=a dictionary containing the key/value pairs required to
        instantiate this SkyElement
        """
        self.fd=dict()
        self.descr='Not yet set'
        self.expr='Not yet set'

        if formdata:
            #pick out what we care about
            for k in self.formkeys:
                self.fd[k]=formdata[k]
                #This will raise a KeyError if any keys of interest
                #are not present. The caller (typically the Sky constructor)
                #will have to deal with this exception

    def __str__(self):
        return str(self.descr)


    def make_synexpr(self):
        """Subclasses must implement this method. It should set both the
        .expr and the .descr attributes
        """
        raise NotImplementedError("Sorry, this must be implemented in the subclass")


class Airglow(SkyElement):
    """Airglow is modeled as a set of narrow emission lines in the UV."""

    #The only input values we care about
    formkeys = ['standard']
    lines = ['el1215a',
             'el1302a',
             'el1356a',
             'el2471a']

    #Normalization constants for each level
    AIRGLOW_LEVELS = {
        'none': None,
        'low': {'el1215a': 0.20,
                'el1302a': 0.01333333333,
                'el1356a': 0.012,
                'el2471a': 0.010},
        'average': 1.0,
        'high': 2.0
        }


    def make_synexpr(self):
        self.descr=self.fd['standard']
        if self.descr == 'none':
            self.expr = None
            return

        elif self.descr == 'low':
            elements=[]
            #Special case, different multipliers for each line
            norm = self.AIRGLOW_LEVELS[self.descr]
            for l in self.lines:
                elements.append("%f*spec(%s)"%(norm[l],MODELS[l]))
            self.expr = ' + '.join(elements)

        else: #All others have a single multiplier applied to all lines
            elements = []
            for l in self.lines:
                elements.append("spec(%s)"%MODELS[l])
            self.expr = '(%s)'%' + '.join(elements)
            if self.AIRGLOW_LEVELS[self.descr] != 1.0:
                self.expr = "%f*(%s)"%(self.AIRGLOW_LEVELS[self.descr], self.expr)


class Earthshine(SkyElement):
    """Earthshine is modelled by a spectrum file."""

    formkeys = ['standard','magnitude','multiplier','choice']
    specfile = "spec(%s)"%MODELS['earthshine']
    # These are scalar multipliers
    EARTHSHINE_LEVELS = {
        'none': None,
        'low': None,
        'average': 0.5,
        'high': 1.0,
        'extremely_high': 2.
        }

    def __init__(self, formdata=None):
        #Call the parent
        SkyElement.__init__(self, formdata)
        #This tells us which type of calculation we're going to do
        self.etype = self.fd['choice']
        self._set_norm()


    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            if self.etype in ('magnitude'):
                self.expr = "rn(%s,band(johnson,v),%f,vegamag)"%(self.specfile,self.norm)
            elif self.etype in ('standard', 'multiplier'):
                self.expr = self.specfile
                if self.norm != 1.0:
                    self.expr = "%f*%s"%(self.norm, self.expr)
            else:
                raise UserInputError(f'Unrecognized normalization option {self.etype}')


    def _set_norm(self):
        if self.etype == 'standard':
            level = self.fd[self.etype]
            self.norm = self.EARTHSHINE_LEVELS[level]
            self.descr = level

        elif self.etype == 'magnitude':
            self.norm = self.fd[self.etype]
            self.descr = 'User Defined (normalized to %f V)'%self.norm

        elif self.etype == 'multiplier':
            self.norm = self.fd[self.etype]
            self.descr = 'User Defined (scaled by %f)'%self.norm

        else:
            raise UserInputError(f'Unrecognized normalization option {self.etype}')


class Helium(SkyElement):
    """Helium is modelled by a spectrum file."""

    formkeys = ['standard']
    specfile = "spec(%s)"%MODELS['helium']

    # These are scalar multipliers
    HELIUM_EMISSION_LEVELS = {
        'none':      None,
        'average':   0.1,
        'high':      0.5,
        'very_high': 1.5,
        }

    def __init__(self, formdata=None):
        SkyElement.__init__(self, formdata)
        self.descr = self.fd['standard']
        self.norm = self.HELIUM_EMISSION_LEVELS[self.descr]

    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            self.expr = self.specfile
            if self.norm != 1.0:
                self.expr = "%f*%s"%(self.norm, self.expr)


class ZodiacalLight(SkyElement):
    """Zodiacal light is modeled by a spectrum file that can be
    normalized in a variety of ways, including based on the angle
    between the sun and the observed coordinates.
    """
    #TODO - Although only a subset of these keys are necessary to
    #construct a valid ZodiacalLight object, at present all are
    #required. We might override the __init__ completely and make it
    #complicated enough to cope with conditionals here.
    formkeys = ['choice','standard','ra','dec',
                'sunattribute', 'sun', 'magnitude', 'multiplier',
                'date', 'month', 'day']
    specfile = "spec(%s)"%MODELS['zodi']
    ZODI_LEVELS = {  #Normalized to Johnson V
        'none': None,
        'low': 23.3,
        'average': 22.7,
        'high': 22.1
    }

    #TODO: handle the posfile in the same way as the other data files
    #  except that some variable expansion has to be done, unlike for
    #  the other files for which that is handled by pysynphot
    #posfile = MODELS['zodipos']

    def __init__(self, formdata=None):
        #Then call the parent
        SkyElement.__init__(self, formdata)
        #This tells us which type of calculation we're going to do
        self.ztype = self.fd['choice']
        self._set_norm()

    def make_synexpr(self):
        if self.norm is None:
            self.expr = None
        else:
            if self.ztype in ('standard', 'magnitude', 'coords'):
                self.expr = "rn(%s,band(johnson,v),%f,vegamag)"%(self.specfile,self.norm)
            elif self.ztype in ('multiplier'):
                self.expr = "%f*%s"%(self.norm,self.specfile)

    def _set_norm(self):
        #There are many ways to specify how the Zodiacal Light spectrum should be normalized
        if self.ztype == 'standard':
            level = self.fd[self.ztype]
            self.norm = self.ZODI_LEVELS[level]
            self.descr = level

        elif self.ztype == 'magnitude':
            self.norm = self.fd[self.ztype]
            self.descr = 'User Defined (normalized to %f V)'%self.norm

        elif self.ztype == 'multiplier':
            self.norm = self.fd[self.ztype]
            self.descr = 'User Defined (scaled by %f)'%self.norm

        elif self.ztype == 'coords':
            zra, zdec = self.fd['ra'], self.fd['dec']
            ra, dec = angles.radec_to_decimal( zra, zdec )

            if self.fd['sunattribute'] == 'date':
                year, month, day = self.fd['date'], self.fd['month'], self.fd['day']
                date = datetime.date(year,month,day)
                helong, helat, sunangle = angles.radec_date_to_helonglat(ra, dec, date)
                self.descr = ('%s %s %s/%s/%s (%f deg.)' % (zra,zdec,month,day,year,sunangle))

            elif self.fd['sunattribute'] == 'sunangle':
                zsun = self.fd['sun']
                # Table is defined from 0-180, but users can put in anything
                sunangle = angles.limit_longitude(float(zsun))
                self.descr = ('%s %s Sun Angle = %f deg.' % (zra, zdec, zsun))
                helong, helat = angles.radec_sunangle_to_helonglat(ra, dec, sunangle)

            else:
                raise UserInputError('Unknown sun specification %s'%self.fd['sunattribute'])
            try:
                self.norm = angles.zodi_lookup(helong, helat)
            except RangeError as e:
                e.addinfo("These values are derived from the RA, Dec, and Date or Helio-ecliptic Longitude. You must either select different values for these quantities, or a different normalization method for the zodiacal light.")
                raise

        else:
            raise UserInputError(f'Unknown normalization type {self.ztype}')


class Sky(object):
    """Class to compute the sky background synphot expression including all relevant SkyElements.
    Presently supports these elements:
        Airglow, Earthshine, and ZodiacalLight.
    Not all the elements are required. Elements will be picked out based on
    the contents of the dictionary that is passed in."""
    default_config = {
        'zodiacallight': {'choice': 'standard', # standard, coords, magnitude, multiplier
                 'standard': 'average',
                 'ra': '00:00:00',
                 'dec': '00:00:00',
                 'sunattribute': 'date', # date, sunangle
                 'date': 2009,
                 'month': 1,
                 'day': 1,
                 'sun': 90.0,
                 'magnitude': 30.0,
                 'multiplier': 1.0},
        'earthshine': {'choice': 'standard',  # standard, magnitude, multiplier
                       'standard': 'average',
                       'multiplier': 1.0, 
                       'magnitude': 30.0},
        'airglow': {'choice': 'standard',
                    'standard': 'average'} # standard
        }

    def __init__(self, config={}, webapp=False, **kwargs):
        self.webapp = webapp

        self.expr = None
        types = recursive_subclasses(SkyElement)
        elements = [t.__name__.lower() for t in types]
        self.sky_map = dict(list(zip(elements, types)))
        self.bg_input = merge_data(self.default_config, config)

    def make_synexpr(self):
        elements = []
        for key in sorted(self.bg_input.keys()):
            sky_elem = self.sky_map[key](self.bg_input[key])
            sky_elem.make_synexpr()
            if sky_elem.expr is not None:
                elements.append(sky_elem.expr)

        if len(elements) > 0:
            self.expr = ' + '.join(elements)




# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import numpy as np
import scipy.integrate as integ
import astropy.units as u

# suppress stsynphot's unnecessary warning (we'll supply our own Vega spectrum)
import warnings
from astropy.utils.exceptions import AstropyUserWarning
# suppress stsynphot's unnecessary warning
warnings.filterwarnings("ignore", "Failed to load Vega spectrum from", AstropyUserWarning)

import synphot as syn
import stsynphot as stsyn
from synphot.models import Empirical1D

from .config import DefaultConfig
from .custom_exceptions import DataConfigurationError, EngineInputError, SynphotError
from .pandeia_warnings import normalization_warning_messages as warning_messages
from .utils import get_key_list, get_dict_from_keys, recursive_subclasses, merge_data
from .instrument_factory import InstrumentFactory
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS, UNIT_MAP, MIN_CLIP

from . import io_utils as io
from . import config as cf

default_refdata_directory = cf.default_refdata_directory

class Normalization(DefaultConfig):

    """
    Class for handling configuration and application of spectrum normalization

    Attributes
    ----------
    type: string
        Type of normalization to perform
    norm_fluxunit: string
        Unit for fluxes used in normalization
    norm_waveunit: string
        Unit for wavelength used for reference wavelength, 'norm_wave'
    norm_wave: float
        Reference wavelength in 'norm_waveunit' at which the spectrum will be scaled for type="at_lambda"
    norm_flux: float
        Reference flux at 'norm_wave' to which spectrum is scaled to match for type="at_lambda"
    bandpass: string
        Specifies how bandpass information is looked up or calculated for normalization
        types 'hst', 'jwst', and 'photsys'.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        How a normalization is configured depends on the defaults, the per-type defaults, and any input parameters.

        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        self.webapp = webapp

        # send webapp=False here since we do the API checking further below
        DefaultConfig.__init__(self, webapp=False, config=config, **kwargs)

        all_config = merge_data(config, dict(**kwargs))

        if not hasattr(self, "type"):
            self.type = self.__class__.__name__.lower().replace('normalize', '')

        if self.type in self.types:
            type_conf = self.types[self.type]
            type_conf = self._load_dict(type_conf, "api", msg=f"No normalization definition for {type(self)}")
            if "defaults" in type_conf:
                type_defaults = type_conf.pop("defaults")
            else:
                type_defaults = {}
            type_conf = merge_data(type_conf, type_defaults, config, dict(**kwargs))
            self.__dict__ = merge_data(self.__dict__, type_conf)

        if hasattr(self,'norm_waveunit'):
            try:
                # If it's not already a quantity, it will be turned into one
                self.norm_wave = syn.units.validate_quantity(self.norm_wave, UNIT_MAP[self.norm_waveunit])
            except Exception as err:
                msg = "Failed to convert reference wavelength units, {}, to pandeia wavelength units, {}: ".format(self.norm_waveunit, PANDEIA_WAVEUNITS, err)
                if self.webapp:
                    msg += "{}".format(type(err))
                else:
                    msg += repr(err)
                raise EngineInputError(value=msg)

        if hasattr(self,'norm_fluxunit'):
            try:
                # If it's not already a quantity, it will be turned into one
                self.norm_flux = syn.units.validate_quantity(self.norm_flux, UNIT_MAP[self.norm_fluxunit])
            except Exception as err:
                msg = "Failed to convert reference flux units, {}, to pandeia flux units, {}: ".format(self.norm_fluxunit, PANDEIA_FLUXUNITS)
                if self.webapp:
                    msg += "{}".format(type(err))
                else:
                    msg += repr(err)
                raise EngineInputError(value=msg)

        # do some sanity checks to make sure inputs make sense
        try:
            self._sanity_checks()
        except AttributeError as e:
            self.warnings['no_sanity_check'] = warning_messages['no_sanity_check'] % (self.__class__.__name__, e)

        # we need to run the API checks here after merging in the normalization type-specific defaults
        if webapp:
            self._api_checks(all_config)

    def _get_config(self):
        """
        Read default configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        ref_dir = os.path.join(default_refdata_directory, "normalization")
        config = io.read_json(os.path.join(ref_dir, "config.json"), raise_except=True)

        config = self._load_dict(config,"api", msg="No normalization API defined.")
        if self.webapp:
            config.pop("defaults")
        else:
            config = self._load_dict(config, "defaults", msg="No normalization defaults defined.")

        return config

    def _sanity_checks(self):
        """
        Make sure norm_fluxunit is supported and type is a configured normalization type
        """
        if self.norm_fluxunit not in self.fluxunits:
            msg = "Unsupported flux unit %s for normalization type %s" % (self.norm_fluxunit, self.type)
            raise EngineInputError(value=msg)

        # Make sure the unit matches what norm_fluxunit is supposed to be
        if self.norm_flux.unit != UNIT_MAP[self.norm_fluxunit]:
            msg = "Unit mismatch: flux is {}, supplied unit is {}".format(self.norm_flux.unit, self.norm_fluxunit)
            raise EngineInputError(value=msg)

        if self.type not in self.types:
            msg = "No configuration data found for normalization type %s" % self.type
            raise DataConfigurationError(value=msg)

class NormalizeAtLambda(Normalization):

    """
    Subclass for handling normalization via a reference flux, self.norm_flux, at a reference wavelength, self.norm_wave.
    """

    def __init__(self, webapp=False, config={}, **kwargs):
        """
        Parameters
        ----------
        webapp: bool
            Determines whether strict API checks should be performed
        config: dict
            Extra configuration information in engine input API dict format
        **kwargs: keyword/value pairs
            Extra configuration information in kwargs format
        """
        Normalization.__init__(self, webapp=webapp, config=config, **kwargs)

        # check sanity of input units
        if self.norm_waveunit not in self.waveunits:
            msg = "Unsupported reference wavelength unit, %s. Supported units are: %s" % (self.norm_waveunit, repr(self.waveunits))
            raise EngineInputError(value=msg)

        if self.norm_wave.unit != UNIT_MAP[self.norm_waveunit]:
            msg = "Unit mismatch: wavelength is {}, supplied unit is {}".format(self.norm_wave.unit, self.norm_waveunit)
            raise EngineInputError(value=msg)

        if self.norm_fluxunit not in self.fluxunits:
            msg = "Unsupported flux unit, %s. Supported units are: %s" % (self.norm_fluxunit, repr(self.fluxunits))
            raise EngineInputError(value=msg)

        # use synphot to convert self.norm_flux to PANDEIA_FLUXUNITS, if necessary.
        try:
            self.norm_flux = syn.units.convert_flux(self.norm_wave, self.norm_flux, PANDEIA_FLUXUNITS)
        except Exception as e:
            msg = "Error using Synphot to convert norm_fluxunit to PANDEIA_FLUXUNITS. "
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)

    def normalize(self, wave, flux):
        """
        Normalize spectrum by scaling it at self.norm_wave.

        Parameters
        ----------
        wave: 1D np.ndarray
            Wavelength vector in 'PANDEIA_WAVEUNITS'
        flux: 1D np.ndarray
            Flux vector in 'PANDEIA_FLUXUNITS' containing the spectrum

        Returns
        -------
        scaled_wave, scaled_flux: 1D np.ndarray
            Wavelength and scaled flux vectors in PANDEIA_WAVEUNITS and PANDEIA_FLUXUNITS
        """
        if self.norm_wave > wave.max() or self.norm_wave < wave.min():
            msg = "Specified normalization wavelength, {:.2f} microns, not within wavelength bounds of spectrum: ({:.2f}, {:.2f} " \
                "microns)".format(self.norm_wave.value, wave.min().to_value(PANDEIA_WAVEUNITS), wave.max().to_value(PANDEIA_WAVEUNITS))
            raise EngineInputError(value=msg)

        spec_flux = np.interp(self.norm_wave, wave, flux)
        if spec_flux == 0.0:
            key = 'normalized_to_zero_flux'
            self.warnings[key] = warning_messages[key]
            scale_factor = 1.0
        else:
            # This operation is NOT astropy unit aware, so we need to be sure both are in PANDEIA_FLUXUNITS.
            # They already should be, but this forces it.
            self.norm_flux = syn.units.validate_quantity(self.norm_flux, PANDEIA_FLUXUNITS)
            spec_flux = syn.units.validate_quantity(spec_flux, PANDEIA_FLUXUNITS)
            scale_factor = self.norm_flux / spec_flux
            normed_flux = np.interp(self.norm_wave, wave, flux/spec_flux)

        scaled_wave = np.copy(wave)
        scaled_flux = flux * scale_factor

        return scaled_wave, syn.units.convert_flux(scaled_wave, scaled_flux, PANDEIA_FLUXUNITS)


class NormalizePhotsys(Normalization):

    """
    Subclass for normalizing a spectrum to a reference flux/magnitude in a filter in a supported photometric system
    (e.g. Cousins, Bessell, SDSS)
    """

    def _get_bandpass(self, *args):
        """
        Parse a self.bandpass of the form <photsys>,<filter> and use the keys to look up a bandpass file
        in self.photsystems.  Use synphot to load the bandpass and return it.

        Returns
        -------
        bp: synphot.spectrum.SpectralElement
            Bandpass throughput data as returned by synphot.spectrum.SpectralElement.from_file() and converted to pandeia
            wavelength units
        """
        try:
            keys = get_key_list(self.bandpass, separator=',')  # should be [photsys, filter]
            bp_filename = get_dict_from_keys(self.bandpasses, keys)['filename']
            bp_path = os.path.join(os.environ.get("PYSYN_CDBS"), "comp", "nonhst", bp_filename)
            bp = syn.spectrum.SpectralElement.from_file(bp_path)
        except Exception as e:
            msg = "Error loading normalization bandpass: %s " % self.bandpass
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataConfigurationError(value=msg)
        return bp

    def normalize(self, wave, flux):
        """
        Normalize a spectrum to a bandpass fetched from self._get_bandpass()

        Parameters
        ----------
        wave: 1D np.ndarray
            Wavelength vector in pandeia wavelength units
        flux: 1D np.ndarray
            Flux vector in 'self.norm_fluxunit' containing the spectrum

        Returns
        -------
        scaled_wave, scaled_flux: 1D np.ndarrays
            Wavelength and scaled flux vectors in pandeia units, microns and mJy
        """

        sp = syn.spectrum.SourceSpectrum(Empirical1D, points=wave, lookup_table=flux)
        bp = self._get_bandpass(wave)
        # Check if the flux is less than 1e-10 mJy
        if integ.trapz(sp(bp.waveset, flux_unit=PANDEIA_FLUXUNITS).value) < MIN_CLIP:
            key = 'normalized_to_zero_flux_bandpass'
            self.warnings[key] = warning_messages[key]
            sp_rn = sp
        else:
            try:
                sp_rn = sp.normalize(self.norm_flux, band=bp, vegaspec=syn.spectrum.SourceSpectrum.from_vega())
            except Exception as e:
                msg = "Error using Synphot to perform bandpass normalization in %s" % str(bp)
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)

        return sp_rn.waveset.to(PANDEIA_WAVEUNITS), sp_rn(sp_rn.waveset, flux_unit=PANDEIA_FLUXUNITS)


class NormalizeObsmode(NormalizePhotsys):

    """
    Subclass for normalizing a spectrum based on a synphot-compatible obsmode string via synphot.spectrum.SpectralElement()
    """

    def _get_bandpass(self, *args):
        """
        Wrap stsynphot.band() to generate a bandpass based on a valid obsmode specification

        Returns
        -------
        bp: synphot.SpectralElement
            synphot bandpass
        """
        if self.webapp and self.bandpass not in self.bandpasses:
            key = "unsupported_normalization_bandpass"
            self.warnings[key] = warning_messages[key] % self.bandpass

        try:
            bp = stsyn.band(str(self.bandpass))
        except Exception as e:
            msg = "Error using Synphot to load bandpass via ObsMode string, %s. " % self.bandpass
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)
        return bp


class NormalizeHst(NormalizeObsmode):

    """
    Largely an alias for NormalizeObsmode since HST normalization bands will be specified as obsmode strings.
    """

    pass


class NormalizeJwst(NormalizePhotsys):

    """
    Subclass for normalizing a spectrum based on a JWST configuration
    """

    def _get_bandpass(self, wave):
        """
        Get JWST instrument, mode, filter from self.bandpass, use that to create a JWSTInstrument instance, and then
        query that for throughput vs wavelength.  Use synphot.SpectralElement to convert the results to
        synphot-compatible form.

        Parameters
        ----------
        wave: 1D np.ndarray
            Wavelength vector onto which JWST bandpass is interpolated

        Returns
        -------
        bp: synphot.SpectralElement
            synphot bandpass converted to pandeia wavelength units.
        """
        keys = get_key_list(self.bandpass, separator=',')
        if len(keys) != 3:
            msg = "JWST bandpass specification must be of the form <instrument>,<mode>,<filter>"
            raise EngineInputError(value=msg)

        instrument, mode, filt = keys
        config = {}
        config['instrument'] = {}
        config['instrument']['instrument'] = instrument
        config['instrument']['mode'] = mode
        config['instrument']['filter'] = filt
        inst = InstrumentFactory(config=config)
        thruput = inst.get_total_eff(wave.to_value(PANDEIA_WAVEUNITS))
        try:
            bp = syn.spectrum.SpectralElement(Empirical1D, points=wave, lookup_table=thruput)
        except Exception as e:
            msg = "Error creating Synphot bandpass from JWST bandpass specification, %s. " % self.bandpass
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise SynphotError(value=msg)
        return bp


class NormalizeNone(Normalization):

    """
    Subclass to make no normalization look like the other methods.
    """

    def normalize(self, wave, flux):
        """
        Implement this method to maintain consistent API and simply return what's given.

        Parameters
        ----------
        wave: 1D np.ndarray
            Wavelength vector in pandeia wavelength units
        flux: 1D np.ndarray
            Flux vector in 'self.norm_fluxunit' containing the spectrum

        Returns
        -------
        wave, flux: 1D np.ndarrays
            Return wave and flux without modification
        """
        return wave, syn.units.convert_flux(wave, flux, PANDEIA_FLUXUNITS)

    def _sanity_checks(self):
        """
        'type' is the only parameter this normalization type accepts. Should never get here
        if NormalizationFactory is used...
        """
        if self.type not in self.types:
            msg = "No configuration data found for normalization type %s" % self.type
            raise DataConfigurationError(value=msg)


def NormalizationFactory(webapp=False, config={}, **kwargs):

    """
    Function to take normalization configuration data and build/return a configured instance of the appropriate
    subclass of Normalization.

    Parameters
    ----------
    webapp: bool
        Switch to toggle strict API checking
    config: dict
        Configuration data in engine API dict format
    **kwargs: keyword/value pairs
        Additional configuration data
    """
    all_config = merge_data(config, dict(**kwargs))
    if 'type' in all_config:
        method = all_config['type'].replace('_', '')
    else:
        msg = "Must specify type of normalization to perform."
        raise EngineInputError(value=msg)

    types = recursive_subclasses(Normalization)
    methods = [t.__name__.lower().replace('normalize', '') for t in types]
    type_map = dict(list(zip(methods, types)))

    if method not in methods:
        msg = "Unsupported or not yet implemented normalization method: %s" % method
        raise EngineInputError(value=msg)
    else:
        cls = type_map[method](webapp=webapp, config=config, **kwargs)
        return cls



import os
import importlib

import pandeia.engine.helpers.bit.pyetc_util as util

class Data(object):
    """
    Interface class to instrument/detector configuration files

    Data class provides the interface between the instrument/detector
    configuration files and the rest of pyetc.

    Parameters
    ----------
    location: str
        location of data files
    instrument: str
        name of instrument
    telescope: str
        name of telescope

    """
    def __init__(self, telescope, instrument, location=__file__):

        self.location = location
        self.location = "/".join(os.path.abspath(location).split('/')[0:-1])

        self.instrument = instrument
        self.telescope = telescope
        self.dconfig = util.stable_dict()
        self.iconfig = util.stable_dict()
        self.custom_config = util.stable_dict()
        self.aperture_fraction = util.stable_dict()
        self.sharpness = util.stable_dict()
        self.extraction_defaults = util.stable_dict()

        self.populate_iconfig()
        self.populate_dconfig()
        self.populate_custom_config()
        self.populate_aperture_fraction()
        self.populate_sharpness()
        self.populate_extraction_defaults()

        # Extract the relevant limits
        self._inst_limits =  importlib.import_module('pandeia.engine.helpers.bit.instruments.' +
                                               telescope + '.' + instrument +
                                               '.' + instrument + '_limits')

    def populate_iconfig(self):
        """
        Populate the iconfig dictionary for use in engine.initialize

        """
        #Read in the instrument config file
        iconfig_name = self.location +'/' + self.instrument + '_config.dat'
        self.iconfig = util.read_dict(iconfig_name)

    def populate_dconfig(self):
        """
        Populate the dconfig dictionary for use in engine.initialize

        """
        dconfig_name = self.location + '/' + 'detector_config.dat'
        self.dconfig = util.read_dict(dconfig_name)

    def populate_custom_config(self):
        """
        Populate the custom_config dictionary for potential use in engine.initialize

        """

        # A custom configuration file is not required. Check for existence
        # and return an empty dictionary if not found.
        try:
            custom_config_name = self.location + '/' + 'custom_config.dat'
            self.custom_config = util.read_dict(custom_config_name)
        except IOError as e:
            self.custom_config = util.stable_dict()

    def populate_aperture_fraction(self):
        """
        Populate the instrument specific aperture fraction table

        """
        #Read in the aperture fraction data file
        try:
            aperture_fraction_name = self.location +'/' + self.instrument + '_aperture_fraction.dat'
            self.aperture_fraction = util.read_dict(aperture_fraction_name)
        except IOError as e:
            self.aperture_fraction = util.stable_dict()

    def populate_sharpness(self):
        """
        Populate the instrument specific sharpness table

        """
        #Read in the sharpness data file
        try:
            sharpness_name = self.location +'/' + self.instrument + '_sharpness.dat'
            self.sharpness = util.read_dict(sharpness_name)
        except IOError as e:
            self.sharpness = util.stable_dict()

    def populate_extraction_defaults(self):
        """
        Populate the extraction default dictionary if available

        """
        #Read in the extraction defaults data file
        try:
            extraction_defaults_name = self.location +'/' + self.instrument + '_extraction_defaults.dat'
            self.extraction_defaults = util.read_dict(extraction_defaults_name)
        except IOError as e:
            self.extraction_defaults = util.stable_dict()

    def get_dconfig(self, detector_name):
        """
        Return a copy of the dconfig dictionary

        """
        return self.dconfig[detector_name].copy()

    def get_iconfig(self):
        """
        Return a copy of the iconfig dictionary

        """
        return self.iconfig.copy()

    def get_custom_config(self):
        """
        Return a copy of the custom_config dictionary

        """
        return self.custom_config.copy()

    def get_aperture_fraction(self):
        """
        Return a copy of the aperture_fraction dictionary

        """
        return self.aperture_fraction.copy()

    def get_sharpness(self):
        """
        Return a copy of the sharpness dictionary

        """
        return self.sharpness.copy()

    def get_extraction_defaults(self):
        """
        Return a copy of the extraction_defaults dictionary

        """
        return self.extraction_defaults.copy()

    def get_relevant_limits(self, mode, detector, filter=''):
        """
        Use the engine.limit_parser interface to
        extract the instrument specific limit information

        """
        from ..engine import limit_parser
        limitdict = limit_parser.extract_limits(mode,
                                                detector,
                                                filter,
                                                self._inst_limits.limit_spec,
                                                self._inst_limits.allsub)

        return limitdict



""" COS-specific functions

From:

    https://stsci-ins.basecamphq.com/projects/9916157-etc-code-reorg/posts/73652016

we note that 3 key assumptions concerning COS mirror-b handling are: 

    1. For COS imaging and imagacq, extended targets are assumed
       to *fill the aperture*.  Thus, they are treated exactly like sky, and
       the dual image would overlap.
    2. For a point target, the secondary image is displaced far enough away
       that it does not overlap.
    3. For a point target, the user (probably) wants the S/N ratio for the
       primary image only, excluding the counts in the secondary image.

See modify_target_or_sky().
"""



from ....engine import instrument, eetable
from ....engine import initialize

from ....engine.cross_collection import ExtractedCrossCollection
from ....engine.exceptions import RangeError, EEDataNotAvailable, InputError
from ....engine.instrument import find_instrument_type
from ....engine.imager import Imager
from ....engine.spectrograph import Spectrograph
from ....engine.detected import DetectedItem, SegmentedDetectedItem
from ....engine.observed import ImageObservedItem
from ....engine.extracted import ExtractedCollection, ExtractedItem
from ....engine.area_over_which import SpecPixAOW, ImageRectangularArea, \
    SpecAreaCollection, SpecSpatialAOW
from ....engine.string_constants import SPECTROSCOPIC, SPECTROSCOPIC_ACQ, RECTANGLE, \
    STRAY_LIGHT, ACQ_WINDOW, SNR_REGION, PIX_REGION, AREA_WHOLE_DETECTOR, TARGET, \
    PIXELS, DARK, DARK_ONLY

from ...telescopes import TELESCOPES
from ....main import tools, util


def get_configuration_data(ei):
    """ Load instrument and detector data from config files.

        This overrides the base function in order to handle
        COS specifics.

    Parameters
    ----------
    ei: dict
      engine input

    Returns
    -------
    iconfig: dict
      dictionary of instrument data
    dconfig: dict
      dictionary of detector data
    custom_config: dict
      dictionary with custom data

    """
    # get baseline configuration data.

    iconfig, dconfig, custom_config = initialize.get_configuration_data(ei)

    detector_name = iconfig['detector_name']

    # in spectroscopy, pixel sizes, dark current, and extraction box
    # sizes depend on several factors: detector, grating, science
    # mode, central wavelength.

    instrument_type = find_instrument_type(iconfig['science_mode'])

    if instrument_type == instrument.spectrograph_type:

        # pick out the correct segaltnames based on grating and cenwave

        altnames = dconfig['segaltnames']
        dconfig['segaltnames'] = altnames.get((ei['grating'],ei['central_wavelength']), None)

        # select correct extraction box sizes

        if not iconfig['acqmode']:
            # Read default scalar values.
            default_spectral_box_width = iconfig[detector_name]['default_spectral_box_width']
            default_spectral_box_height = iconfig[detector_name]['default_spectral_box_height']

            # Read per-cenwave values.
            spectral_box_width = iconfig[detector_name]['spectral_box_width']
            spectral_box_height = iconfig[detector_name]['spectral_box_height']

            # Initialize with scalar defaults.
            iconfig['box_height'] = default_spectral_box_height
            iconfig['box_width'] = default_spectral_box_width

            # In the FUV, defaults may be superseded by finer-grained dependencies.
            if isinstance(spectral_box_height, dict):

                # as per config file, heights may have per-grating defaults that
                # apply to all cenwaves for that given grating. Or, heights may
                # depend on both grating and cenwave. Currently there are no cases
                # where the extraction height doesn't depend on anything. The logic
                # here is based on that, but may have to change in the future if
                # the data structures in the config file are changed.
                spectral_box_height_defaults = iconfig[detector_name]['spectral_box_height_defaults']

                default_heights_for_grating = spectral_box_height_defaults.get(iconfig['spec_element'], default_spectral_box_height)
                heights_for_grating = spectral_box_height.get(iconfig['spec_element'], default_heights_for_grating)
                if isinstance(heights_for_grating, dict):
                    iconfig['box_height'] = heights_for_grating.get(str(ei["central_wavelength"]), default_heights_for_grating)

                # as per config file, widths always depend on both
                # grating and  cenwave, or on none of them.
                widths_for_grating = spectral_box_width.get(iconfig['spec_element'], None)
                if widths_for_grating:
                    iconfig['box_width'] = widths_for_grating.get(str(ei["central_wavelength"]), default_spectral_box_width)

        else:
            custom_config['acq_mode'] = ei['extraction']['acq_mode']
            custom_config['stripe'] = ei['extraction']['stripe'].lower()

    if instrument_type == instrument.spectrograph_type and type(dconfig['pixel_height']) == type(util.stable_dict()):

        pixel_height = dconfig['pixel_height'][iconfig['science_mode']][iconfig['spec_element']]
        pixel_width = dconfig['pixel_width'][iconfig['science_mode']][iconfig['spec_element']]
        dark_rate = dconfig['dark_current_rate'][iconfig['science_mode']]

        dconfig['pixel_height'] = pixel_height
        dconfig['pixel_width'] = pixel_width
        dconfig['dark_current_rate'] = dark_rate

    spectral_resolution = iconfig[detector_name]['spectral_resolution']

    if instrument_type == instrument.spectrograph_type and type(spectral_resolution) == type(util.stable_dict()):
        key = (ei["grating"],ei["central_wavelength"])
        if key in list(spectral_resolution.keys()):
            iconfig[detector_name]['pixels_per_resel'] = spectral_resolution[key]
        else:
            iconfig[detector_name]['pixels_per_resel'] =  spectral_resolution["default"]
    else:
        iconfig[detector_name]['pixels_per_resel'] = spectral_resolution

    # in imaging, the mirror type (A or B) must be configured.
    if instrument_type == instrument.imager_type:
        iconfig['spec_element'] = ei['mirror']

    return iconfig, dconfig, custom_config


def get_detector_tables(instrument_name, detector_name, science_mode, spec_element):
    """ Gets enclosed energy tables for COS spectroscopy.

        COS FUV has a highly variable PSF across the FOV,
        thus uses non-standard tables which depend not only
        on the primary information (grating, detector, science
        mode), but also on central wavelength and detector
        segment.

        COS NUV, on the other hand (in spec acq mode) has
        the table selected on the basis of segment only.

    Parameters
    ----------
    instrument_name: string
      the instrument name
    detector_name: string
      the detector name. Not used.
    science_mode: string
      the science mode.
    spec_element: str
      the 'disperser' entry in the engine input dict.

    Returns
    -------
    result: dict
        Dictionary with the enclosed energy tables associated to this specific disperser.

    """
    # NOTE: the old code uses a generic table reading infrastructure to select the
    # right table, given the complex selection criteria required by spec and spec acq
    # modes. I found the code hard to understand without actually running it in a
    # debugger. I adopted here a more explicit approach, with hard-coded selection criteria.
    # I found this way easier to understand on a first reading. It also doesn't
    # compromise anything, since it is contained in the COS module and is so COS-specific.

    # Mostly, duplicate the generic code to
    # ingest the entire set of tables.

    instconfig = tools.find_instrument(TELESCOPES[instrument_name], instrument_name)
    APDATA = instconfig.data.get_aperture_fraction()

    tables = APDATA['tables']

    # But then get tables using the finer-grained input info.

    result = util.stable_dict()
    for table in tables:
        if science_mode == SPECTROSCOPIC:
            segment = table['segment']

            # In FUV spec mode, tables are selected based
            # on grating, segment, and eventually cenwave.
            if table['science_mode'] == SPECTROSCOPIC and \
               table['detector']     == 'fuv' and \
               detector_name         == 'fuv' and \
               table['grating']      == spec_element:

                # segments may have a single table that
                # applies to all cenwaves. Use a wild card
                # character in that case.
                cenwave = table.get('cenwave','#')

                # We code the extra information in the dict key so multiple
                # tables can be passed around on a single calculation.
                result[table['aperture_shape'] + '_' + str(cenwave) + '_' + str(segment)] = table

            # In NUV spec acq mode, tables are selected based
            # on segment (stripe) only.
            if table['science_mode'] == SPECTROSCOPIC and \
                detector_name        == 'nuv' and \
                 table['detector']   == 'nuv':

                result[table['aperture_shape'] + '_#_' + str(segment)] = table

        else:
            # Or, just fall back again to the base code.
            result = eetable.get_detector_tables(instrument_name, detector_name, science_mode)

    return result


def modify_configuration(instrument, custom_config):
    """ The instrument instance is populated with
        custom configuration attributes.

        Configuration step in COS must pick up extra information
        from the configuration files. In imaging mode, the mirror
        B multiplicative factor.

        For a segmented detector in spec mode, the initialization
        code cannot tell apart EE tables that apply to the same
        shape and grating, but are differentiated by central
        wavelength and segment (or just segment, being valid for
        all non-explicit central wavelengths). This method
        refines the choice by decreasing the availability of EE
        tables to just one per segment.

    Parameters
    ----------
    instrument: Imager or Spectrograph
      Instance being configured
    custom_config: dictionary
      custom configuration info

    """
    if isinstance(instrument, Spectrograph):

        #acq windows and limits
        if instrument.acqmode:
            # the acq config dictionaries for FUV and NUV have different keys
            # and different nesting levels.
            detector_regions = {'fuv': 'windows', 'nuv': 'specacq_extraction_height'}

            detector_name = instrument.detector.name
            spectral_element = instrument.spectral_element
            custom_dict_key = detector_regions[detector_name]

            instrument.custom.windows = custom_config[detector_name][custom_dict_key][spectral_element]

            if detector_name == 'fuv':
                cenwave = instrument.central_wavelength
                instrument.custom.windows = instrument.custom.windows[cenwave]

            #1294 - keep only limits that apply to spectroscopic acq mode. The Limit class
            # and associated code (limit parser) cannot tell apart spectroscopic mode from
            # spectroscopic target acq mode. In both modes, science_mode is set to 'spectroscopic'.
            # The class should be able to access the acqmode flag to properly process limits that
            # apply to one or another mode, but that is not implemented yet. Since this problem
            # only affects COS, we apply here a instrument-specific solution. See #1294 for
            # more info.
            new_limits = util.stable_dict()
            for limit_name in list(instrument.limits.keys()):
                limit = instrument.limits[limit_name]
                if SPECTROSCOPIC_ACQ in limit._defn['mode']:
                    new_limits[limit_name] = limit

            instrument.limits = new_limits

        # EE tables can be indexed by central wavelength and
        # segment. Here is an opportunity to prune the dictionary
        # and leave just the tables that apply to a given central
        # wavelength.
        input_eetables = instrument._eetables
        if isinstance(instrument._eetables, dict):
            result = util.stable_dict()
            # First, look for explicit match with cenwave.
            for key in list(input_eetables.keys()):
                name, under, leftover = key.partition('_')
                cenwave, under, segname = leftover.partition('_')

                if cenwave == str(instrument.central_wavelength):
                    table = input_eetables[key]
                    table.name = name
                    result[segname] = table

            # If no explicit match was found, look for the
            # generic EE table (keyed with a '#', as coded
            # by the get_detector_tables method.
            if not len(result):
                for key in list(input_eetables.keys()):
                    name, a, leftover = key.partition('_')
                    cenwave, a, segname = leftover.partition('_')

                    if cenwave == '#':
                        table = input_eetables[key]
                        table.name = name
                        result[segname] = table

            # If still no match, that's an error.
            if not len(result):
                raise EEDataNotAvailable("For %d ."%instrument.central_wavelength)

            instrument._eetables = result

    elif isinstance(instrument, Imager):
        # mirror B
        instrument.custom.mirrorb_correction = custom_config[instrument.detector.name]['mirrorb_correction']
    else:
        raise TypeError("Unrecognized instrument of type %s"%type(instrument))


def _make_snr_region_collection(instrument, *args, **kwargs):
    """
    In COS spec acq mode we have one SNR region per window;
    we store them in a SpecAreaCollection instance.

    """
    rgn_collection = SpecAreaCollection(SNR_REGION)
    pix_rgn_collection = SpecAreaCollection(PIX_REGION)

    windows = instrument.custom.windows
    specacq_extraction_height = instrument.areas_of_interest['area_over_whole_detector']._collection

    for wname in windows:

        # For FUV, both the extraction height and the wavelength
        # range of each AOW are taken from the custom configuration
        # file. The AOWs are designed in that way to keep out
        # contamination from geocoronal lines. This information
        # is not present in the throughput of the instrument, thus
        # it has to be coded explicitly in the ETC config files.
        if instrument.detector.name == 'fuv':
            start = windows[wname][0][0]
            stop = windows[wname][0][1]
            height = windows[wname][0][2]

        # For NUV, one can use directly the stripes defined by
        # the instrument throughput. The config file can have
        # just the AOW heights.
        if instrument.detector.name == 'nuv':
            height = windows[wname][0]
            start = specacq_extraction_height[wname]._waverange[0]
            stop = specacq_extraction_height[wname]._waverange[1]

        rgn = SpecSpatialAOW(wname,
                             height = height,
                             height_units = PIXELS,
                             waverange = [start,stop],
                             pixrange = [0, instrument.detector.xpixels-1],
                             pixel_angheight = instrument.detector.pixel_angheight,
                             pixel_angwidth = instrument.detector.pixel_angwidth)

        # 1-pix wide region positioned midpoint in the waverange.
        pix_rgn = SpecPixAOW(wname, height, args[0][1],
                             (start + stop) / 2.,
                             1,
                             instrument.detector.pixel_angheight,
                             instrument.detector.pixel_angwidth)

        rgn_collection.add(rgn)
        pix_rgn_collection.add(pix_rgn)

    return rgn_collection, pix_rgn_collection


def make_snr_region(instrument, *args, **kwargs):
    """
    COS uses specialized SNR regions for target acquisition modes,
    both image and spec. The SNR region for spec mode might also
    need special configuration, depending on which segment it's
    going to perform the extraction.

    """
    if isinstance(instrument, Spectrograph):

        # Extraction height and width are custom-set
        # by the initialization module in the
        # segmented detector case. Extraction height
        # can be segment(obswave)-dependent as well.

        segments = instrument.detector.segments
        if segments and len(segments) > 0:

            width = instrument.custom.box_width

            obswave = args[2]

            if isinstance(instrument.custom.box_height, dict):
                height = None
                for segname in segments:
                    if obswave in segments[segname]:
                        height = instrument.custom.box_height[segname]

                if not height:
                    msg = ""
                    for segname in segments:
                        msg += str(segments[segname].range)

                    msg = "Obswave %f not in any segment (%s)" % (obswave, msg)
                    raise RangeError(msg)

            else:
                height = instrument.custom.box_height

            if instrument.acqmode:
                # Spec acq mode has a collection
                # of SNR regions, one per window.
                rgn, pix_rgn = _make_snr_region_collection(instrument, args, kwargs)

            else:
                # Spec mode has a single SNR region.
                rgn = SpecPixAOW(SNR_REGION, height, args[1], obswave,
                                 width,
                                 instrument.detector.pixel_angheight,
                                 instrument.detector.pixel_angwidth)
                pix_rgn = SpecPixAOW(PIX_REGION, height, args[1], obswave,
                                     1,
                                     instrument.detector.pixel_angheight,
                                     instrument.detector.pixel_angwidth)

            instrument.areas_of_interest[PIX_REGION] = pix_rgn

            return rgn

        if instrument.acqmode:
            rgn = instrument.areas_of_interest[ACQ_WINDOW]
        else:
            rgn = None
    else:

        # image acq SEARCH mode requires a rectangular extraction region

        if args[0] == RECTANGLE:

            width  = args[1][0]
            height = args[1][1]
            units  = args[2]

            arcsec_per_pixel_height = instrument.detector.pixel_angheight
            arcsec_per_pixel_width = instrument.detector.pixel_angwidth

            rgn = ImageRectangularArea(SNR_REGION,
                                       width,
                                       height,
                                       units,
                                       arcsec_per_pixel_width,
                                       arcsec_per_pixel_height)
        else:
            rgn = None

    return rgn


def detect_dark(instrument, dark):
    """ Returns the value of the dark current per pixel.

    Parameters
    ----------
    instrument: Spectrograph
        configured instance
    dark: float or dict
      dark rate value as a scalar, or dictionary with
      dark rate per segment.

    Returns
    -------
    result: DetectedItem or DetectedCollection
      Dark current rate in counts per second per pixel.

    Notes
    -----
    """
    if isinstance(dark, dict):
        result = SegmentedDetectedItem(DARK, dark, instrument)
    else:
        result = DetectedItem(DARK, dark, instrument.detector.shape)

    return result


def modify_areas_of_interest(instrument):
    """ Modify the dictionary of areas other than the snr_region.

    The .areas_of_interest is a dictionary of {name:AreaOverWhich} that defines
    all areas to be used in a calculation other than the area to be used for the
    SNR or Time calculation itself. Usually, these will be pertinent to the
    feasibility/diagnostic stream.

    Certain area names are predefined in engines/string_constants.py and are
    used by the code.

    Other strings are permitted to support custom operations.
    TODO: define exactly the conditions for these strings.

    For COS specacq mode, we must add the correct windows to this dictionary.

    Parameters
    ----------
    instrument: Imager or Spectrograph
       the configured COS instance

    Returns
    -------
    None

    Side effects
    ------------
    Modifies instrument.areas_of_interest
    """

    # If we are in specacq mode, then the ACQ_WINDOW entry must be added to the
    # areas of interest.

    # The windows are specified as discontiguous arrays in order to exclude
    # geocoronal lines that will mess up the acquisition. But from the ETC's
    # point of view, we need to have one window collection per segment/stripe,
    # because that's the level of detail at which they are reported out.
    # Thus we will make a collection of collections.

    if isinstance(instrument, Spectrograph) and instrument.acqmode:

        # Segments must be initialized here because
        # there is no guarantee at this point that
        # the instrument is already fully configured.
        if not instrument.detector.segments:
            instrument.detector.create_segments(instrument.target_point_bandpass)

        window_collection = SpecAreaCollection(ACQ_WINDOW)

        pixel_angheight, pixel_angwidth = (instrument.detector.pixel_angheight,
                                           instrument.detector.pixel_angwidth)
        for name, value in list(instrument.custom.windows.items()):
            segment_collection = SpecAreaCollection(name)

            for k,subrange in enumerate(value):

                if instrument.detector.name == 'fuv':
                    # Windows are named by segment/stripe, then by enumeration.
                    subname = name+str(k)
                    lowave, hiwave, height = subrange

                # NUV has its windows defined by the bandpasses (that is,
                # segment ranges) instead of the window range definitions
                # in the config file.
                if instrument.detector.name == 'nuv':
                    subname = name
                    height = subrange
                    lowave = instrument.detector.segments[name].range[0]
                    hiwave = instrument.detector.segments[name].range[1]

                width = instrument.target_point_bandpass.pixel_range([lowave, hiwave])

                window = SpecSpatialAOW(subname, height, PIXELS,
                                        waverange =(lowave, hiwave),
                                        pixrange = [0, width],
                                        pixel_angheight = pixel_angheight,
                                        pixel_angwidth = pixel_angwidth
                )

                # Add each window to the collection for the segment
                segment_collection.add(window)

            # Add each segment collection to the overall window collection
            window_collection.add(segment_collection)

        # All done - now add that to the areas_of_interest
        instrument.areas_of_interest[ACQ_WINDOW] = window_collection

    # In spec mode the NUV detector is divided in 3 optical stripes. The
    # config file just tells us the size of the entire detector; nowhere
    # to be seen is the height and width of the actual areas occupied by
    # the stripes. The generic code builds the AREA_WHOLE_DETECTOR using
    # the detector size as per config file and thus overestimates the dark
    # rate by a factor 3, since it builds actually a collection of 3 areas,
    # to handle the multiple stripes. Each area is sized as the whole
    # detector, which is the default behavior of the generic code.
    #
    # Here we override the generic code to build extraction regions that
    # are 1/3 of the height of the entire detector. Thus the 3 regions put
    # together side by side cover the entire detector only once and the
    # collection can then deliver the right dark rate for the entire detector.
    #
    # Target and sky are geometrically limited by the  circular aperture
    # and thus are no factor in this context.
    if isinstance(instrument, Spectrograph) and instrument.detector.name == 'nuv':
        coll = instrument.areas_of_interest[AREA_WHOLE_DETECTOR]
        areas = []
        for area in coll:
            areas.append(SpecSpatialAOW(area.name,
                                        area._linear_height / 3,
                                        area._height_units,
                                        area._waverange,
                                        area._pixrange,
                                        area._pixel_angheight,
                                        area._pixel_angwidth))

        # There are cases where only two spectral ranges are available. The loop
        # above will then build two AOWs, which is not enough to cover the entire
        # detector. Here we build a third AOW of a special class that ignores the
        # illumination components that shine on the detector, such as target and
        # sky. That is, this special AOW only accounts for quantities originating
        # on the detector itself, such as dark rate.
        #
        # We copy the AOW parameters from the first AOW built by the loop above.
        # Note that only the pixel-based parameters are relevant here. Thus we
        # use as _waverange a 2-element, zero-valued sequence (a None value causes
        # the initialization machinery in the base class to complain).
        if len(areas) == 2:
            areas.append(_COS_NonIlluminated_SpecSpatialAOW(DARK_ONLY,
                                                areas[0]._linear_height,
                                                areas[0]._height_units,
                                                [0, 0],
                                                areas[0]._pixrange,
                                                areas[0]._pixel_angheight,
                                                areas[0]._pixel_angwidth))

        name = instrument.areas_of_interest[AREA_WHOLE_DETECTOR].name

        instrument.areas_of_interest[AREA_WHOLE_DETECTOR] = SpecAreaCollection(name, contents=areas)


def modify_target_or_sky(instrument, observed_target, observed_sky,
                         obswave=None):
    """  Modifies target and sky observed items according to
         instrument-specific rules.

    Imaging:

    In this case, if the instrument is configured in imaging mode
    with mirrorb, and if the target is extended, it is modified to
    include the light from the secondary image.

    If the target is a point, then we only modify here in the case
    of the area_over_whole_detector, for the feasibility thread.

    The sky is unmodified: the secondary image due to the sky from
    mirror B is applied as a separate stray light effect in the
    stray_light function.

    Spectroscopy:

    In case of a segmented detector, multiple EE tables exist, one
    per segment. Only at this point we can decide which one to pick,
    based on the obswave.

    Parameters
    ----------
    observed_target: ObservedItem
       the target to be modified
    observed_sky: ObservedItem
       the sky to be modified
    obswave: float, optional
       the obswave. Required for segmented detectors.

    Returns
    -------
    tuple with ObservedItems
      new instances of ObservedItem with the modified target and/or
      sky, or references to the input, if no modification took place.

    """
    if instrument.spectral_element == 'mirrorb':
        modified_tgt_obs = \
            observed_target.observation * instrument.custom.mirrorb_correction
        # enter this modified obs. object into the special_target_obs_dict
        if not hasattr(observed_target, 'special_target_obs_dict'):
            observed_target.special_target_obs_dict = util.stable_dict()
        # point targets and extended targets
        observed_target.special_target_obs_dict['area_over_whole_detector'] = \
            modified_tgt_obs
        # extended targets
        if observed_target.visible_shape.angarea > 0.0:
            observed_target.special_target_obs_dict['snr_region'] = \
                modified_tgt_obs
            observed_target.special_target_obs_dict['area_over_one_pixel'] = \
                modified_tgt_obs

    # spectrograph in non-acq mode with segmented detector:
    # pick up EE table for segment that contains obswave.
    if isinstance(instrument, Spectrograph) and \
        instrument.detector.segments and \
        not instrument.acqmode and \
        len(instrument.detector.segments) > 0 and \
        hasattr(observed_target, '_eetables'):

        if not obswave:
            raise InputError("Obswave must be specified for COS spectroscopy.")

        # select EE table associated with the chosen segment.
        for key in list(instrument.detector.segments.keys()):
            if obswave in instrument.detector.segments[key]:
                table = observed_target._eetables[key]
                observed_target._eetables[table.name] = table

                return observed_target, observed_sky
            else:
                continue

        raise EEDataNotAvailable("Obswave %f not in detector segments."%obswave)

    return observed_target, observed_sky


def _init_custom_config(instr, instrument_type, iconfig, dconfig, custom_config):
    """
    Add custom configuration data to the instrument.custom Struct Object

    """

    # Populate for all COS modes
    instr.custom.half_buffer = dconfig['half_buffer']

    # Only populate for COS imaging modes
    if instrument_type == instrument.imager_type:
        m = (iconfig['spec_element'])[:-1]
        m = m[0].swapcase() + m[1:]
        l = (iconfig['spec_element'])[-1]
        l = l.swapcase()
        mirror = m + ' ' + l
        instr.custom.output.mirror = mirror
        # Populate filter bandpass limits into custom config for use in calculating
        # filter leaks (#983)
        instr.custom.bandpass_limits = iconfig['bandpass_limits']

    # Extraction region dimensions may be wavelength-dependent
    if instrument_type == instrument.spectrograph_type:
        instr.custom.box_height = iconfig.get('box_height', None)
        instr.custom.box_width = iconfig.get('box_width', None)

    # Only for spec acq mode.
    if instrument_type == instrument.spectrograph_type and instr.acqmode:
        instr.custom.acqmode = custom_config['acq_mode']
        instr.custom.stripe = custom_config['stripe']
    return instr


def compute_buffer_time(instrument, detector_total_rate):
    """
    Implement a custom COS buffer time calculation

    """
    result = instrument.custom.half_buffer / detector_total_rate
    instrument.custom.output.buffer_time = result

    instrument.custom.output.segmented_buffer_time = {}
    if hasattr(instrument.detector, 'segmented_detector_total_rate'):
        for segname in instrument.detector.segmented_detector_total_rate:
            bt = instrument.custom.half_buffer / instrument.detector.segmented_detector_total_rate[segname]
            instrument.custom.output.segmented_buffer_time[segname] = bt

    return instrument


def stray_light(instrument, target, sky):
    """ Get stray light component.

    Parameters
    ----------
    instrument: Imager
       In COS, stray light corrections apply only to imaging modes.
       If a Spectrograph instance is provided instead, the method
       returns silently as a no-op.
    target: ObservedItem
      the target - not used by COS
    sky: ObservedItem or ObservedCollection
      the sky

    Returns
    -------
    result: ObservedItem, or None.
      If no stray light exists, returns None

    In COS, the mirror B configuration produces a secondary image that
    results in the light coming from the sky to be boosted by a fixed
    multiplicative factor. See documentation here:

        http://www.stsci.edu/*hst*/*cos*/documents/isrs/ISR2010_10.pdf

        http://www.stsci.edu/hst/cos/documents/handbooks/current/ch08.Acquisitions05.html#430115

    Since we are defining the stray light component as an additive
    component to the sky, we subtract 1.0 from the mirror B multiplicative
    factor before applying it to build the stray light ObservedItem
    instance that gets returned by this method.

    """
    if instrument.spectral_element != 'mirrorb':
        return None

    stray_light_observation = sky.observation * (instrument.custom.mirrorb_correction - 1.0)

    qyc_function = instrument.detector.get_qyc_function()

    stray_light = ImageObservedItem(STRAY_LIGHT, stray_light_observation,
                                    sky.visible_shape,
                                    qyc_function,
                                    warnings=sky._warnings)

    return stray_light

#we should let the custom
#stray_light routine compute it, and return an ObservedCollection named
#STRAY_LIGHT that may have SECONDARY_TARGET_IMAGE and SECONDARY_SKY_IMAGE
#in it. I think this would be a *lot* clearer than the current "mirror
#correction" implementation in the code.
#
#It seems odd to me that the secondary target image is used as part of
#the extracted target - I would think it would be noise, not signal. Do
#you understand it?
#
#But anyway, we could then have a modify_extracted_values custom function
#that could do this modification.


def modify_one_pixel(instrument, collection):
    """ Modifies an ExtractedCollection to
        perform brightest pixel calculations
        on a per-segment basis.

    Parameters
    ----------
    collection: ExtractedCollection
      the collection to be modified

    Returns
    -------
    result: COS_ExtractedCrossCollection
      a decorated version of the input collection,
      with the ability to build a dict with
      brightest pixel info, one entry per
      segment.

    """
    result = COS_ExtractedCrossCollection(instrument, collection)
    return result


class COS_ExtractedCrossCollection(ExtractedCrossCollection):
    """ Wrapper (Decorator) that modifies an ExtractedCollection
        with the ability to perform and report brightest pixel
        calculations on a per-segment basis.

        The resulting instance is a decorated version of the
        input collection, with the ability to build a dict with
        brightest pixel info (flux, wavelength, extraction fraction),
        one entry per segment.

    Parameters
    ----------
    instrument: Spectrograph
       the instrument
    contents: ExtractedCollection
       the collection to be augmented with a per-segment
       ability to compute the brightest pixel.

    """
    def __init__(self, instrument, contents):

        ExtractedCrossCollection.__init__(self, PIX_REGION, contents._collection)

        self._segments = instrument.detector.segments

    def brightest_pixel(self):
        """ Computes value and wavelength of brightest pixel.
            Separate values are computed for each segment.

        Parameters
        ----------
        None

        Returns
        -------
        result: dict with ExtractedItem instances
          extracted items with the (scalar) brightest pixel
          value and its associated wavelength (and eventual
          extraction fraction, for point targets), one entry
          per segment.

        """
        result = util.stable_dict()

        for segment_name in self._segments:
            keys = self._get_item_keys(segment_name)
            contents = []
            for key in keys:
                contents.append(self._collection[key])

            collection = ExtractedCollection(segment_name, contents)

            wave, flux = collection.spectrum().arrays()
            target = collection[TARGET]
            if target._eefrac_used:
                eefrac = target._eefrac_used._throughputtable
            else:
                eefrac = None

            bflux, bwave, eefrac= self._segments[segment_name].get_brightest_pixel(wave, flux, eefrac)

            result[segment_name] = ExtractedItem(segment_name,
                                                 bflux,
                                                 None,
                                                 eefrac_used = eefrac,
                                                 wavelength = bwave)

        return result

class _COS_NonIlluminated_SpecSpatialAOW(SpecSpatialAOW):
    """ Overrides base class in order to support COS NUV
        2-stripe entire detector extractions.

        This private class is specifically designed to solve
        the COS NUV 2-stripe problem when computing the dark
        rate over the entire detector. Use in any other
        scenario will most surely cause errors.

    """
    def countrate(self, observation):
        """ Compute the countrate over the region.

        Overrides base class to return zero. The countrate
        over a non-illuminated stripe is zero by definition.

        Parameters
        ----------
        observation: ignored

        Returns
        -------
        result: 0.0, float constant
          counts per second

        """
        return 0.0



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from .telescope import Telescope
from .instrument import Instrument
from .custom_exceptions import EngineInputError
from .pandeia_warnings import instrument_warning_messages

class Roman(Telescope):

    """
    Currently a dummy class for directory/file discovery, but could eventually contain Roman-specific methods
    """
    pass


class RomanInstrument(Instrument):

    """
    Generic Roman Instrument class
    """

    def __init__(self, mode=None, config={}, **kwargs):
        telescope = Roman()
        # these are the required sections and need to be passed via API in webapp mode
        self.instrument_pars = {}
        self.instrument_pars['detector'] = ["nexp", "nresultants", "ma_table_name", "subarray"]
        self.instrument_pars['instrument'] = ["aperture", "disperser", "filter", "instrument", "mode"]
        self.api_parameters = list(self.instrument_pars.keys())

        # these are required for calculation, but ok to live with config file defaults
        self.api_ignore = ['dynamic_scene', 'max_scene_size', 'scene_size']

        Instrument.__init__(self, telescope=telescope, mode=mode, config=config, **kwargs)

    def read_detector(self):
        """Read in the detector keyword from the aperture parameter in the config json file.
           Put the detector keyword in self.instrument['detector']."""
        self.instrument['detector'] = self.aperture_config[self.get_aperture()]['detector']


class WFI(RomanInstrument):

    """
    Currently, the Roman WFI requires only one method beyond those provided by the generic Instrument class
    """
    def __init__(self, mode=None, config={}, webapp=False, **kwargs):

        # We will stick with our own terminology, but allow for someone simply using "optical_element"
        # If present, it supercedes filter and disperser.
        if "optical_element" in config["instrument"]:
            if config["instrument"]["mode"] == "imaging":
                config["instrument"]["filter"] = config["instrument"]["optical_element"]
                config["instrument"]["disperser"] = None
            elif config["instrument"]["mode"] == "spectroscopy":
                config["instrument"]["disperser"] = config["instrument"]["optical_element"]
                config["instrument"]["filter"] = None

        RomanInstrument.__init__(self, mode=mode, config=config, webapp=webapp, **kwargs)
        self._loadpsfs()

        if config["detector"]["nresultants"] == -1:
            key = "roman_max_resultants"
            self.warnings[key] = instrument_warning_messages[key].format(self.the_detector.exposure_spec.max_resultants)

        if config["instrument"]["mode"] == "imaging" and config["instrument"]["filter"] not in self.filters:
            raise EngineInputError(f'Invalid filter: {config["instrument"]["filter"]}')
        elif config["instrument"]["mode"] == "spectroscopy" and config["instrument"]["disperser"] not in self.dispersers:
            raise EngineInputError(f'Invalid disperser: {config["instrument"]["disperser"]}')

        # Mixing the MA tables will not be allowed, so it should error.
        if self.instrument["mode"] == "imaging" and self.the_detector.exposure_spec.ma_table["observing_mode"] != "WIM":
            raise EngineInputError(f"Invalid MA table {self.the_detector.exposure_spec.ma_table_name} for imaging.")
        if self.instrument["mode"] == "spectroscopy" and self.the_detector.exposure_spec.ma_table["observing_mode"] != "WSM":
            raise EngineInputError(f"Invalid MA table {self.the_detector.exposure_spec.ma_table_name} for spectroscopy.")

    def _loadpsfs(self):
        """
        Short-wavelength filters need PSFs with the skinny pupil mask.
        Long-wavelength filters (just f184 and f213 for now) need the wide pupil mask.
        The grism and prism both have their own pupil masks.
        Because the ranges overlap, we need to switch between PSF libraries.
        The mask is attached to the actual glass, so selecting by filter seems appropriate
        """
        if self.instrument['mode'] == 'imaging':
            psf_key = self.instrument['filter']
        elif self.instrument['mode'] == 'spectroscopy':
            psf_key = self.instrument['disperser']
        else:
            message = "Invalid mode specification: {}".format(str(self.instrument['mode']))
            raise EngineInputError(value=message)

        self.psf_library = self._load_psf_library(psf_key)


class IFU(RomanInstrument):

    """
    Currently the Roman IFU (deprecated) requires no extra methods beyond those provided by the generic Instrument class
    """
    pass



#!/usr/bin/env python

'''
This code was copied from the straylight area of the BMG svn repo (rev 352)

One minor modification was made:
    1) convert ra and dec from radians
'''

import numpy as np
import os

from pandeia.engine.custom_exceptions import StraylightPositionError
from pandeia.engine.custom_exceptions import StraylightDataError
from pandeia.engine.custom_exceptions import DatelessBGError

import pandeia.engine.helpers.background.jwst.accessor_globals as glb

from pandeia.engine.helpers.background.jwst.mod_healpix_func import ang2pix_ring

SL_NWAVE = len(glb.wavelist)
THERMAL_FNAME = os.environ['THERMAL_FNAME']


def return_index(ra, dec):  # This takes radians!
    return ang2pix_ring(glb.nside, ra, dec)


nonzodi_pix_dtype = np.dtype(
    [
        ('pix_ra', 'f8'),
        ('pix_dec', 'f8'),
        ('upos', [('x', 'f8'), ('y', 'f8'), ('z', 'f8')]),
        ('nonzodi_bg', ('f8', SL_NWAVE)),
        ('iday_index', ('i4', glb.NUM_DAYS))
    ]
)

zodi_sl_dtype = np.dtype(
    [
        ('zodi_bg', ('f8', SL_NWAVE)),
        ('stray_light_bg', ('f8', SL_NWAVE))
    ]
)

dateless_index_dtype = np.dtype([('dateless_index', ('i4', 3))])


def get_sl(ra, dec, mjd2000, ra_dec_str, date_str):
    """
    Provides equivalent straylight background spectra for pointing and mjd2000.

    Parameters
    ----------
    ra : double
        Right Ascension [degrees]
    dec : double
        declination [degrees]
    mjd2000 : int
        Modified Julian Day 2000
    ra_dec_str: str
        String representation of RA and DEC (for sexigesimal in messages)
    date_str: str
        String representation of date (for Gregorian calendar dates)

    Returns
    -------
    tuple of:
    wave : numpy.ndarray
        Wavelengths of background values [microns]
    stray_light_bg : numpy.ndarray
        Equivalent in field background from the scattered zodi, ism, cbi, stellar.

    If the target is not in the Field of Regard on iday, then an exception is raised.
    """

    # Cache is based on DOY 2020 so convert MJD2000 to DOY.
    # 51544 is Jan 1, 2000
    iday = int(float(mjd2000) % 365.25) + 1

    # Find the HEALPix number and the subdirectory
    ipix = ang2pix_ring(glb.nside, ra*glb.D2R, dec*glb.D2R)
    ipix_dir = ipix // 100

    # Set up the numpy arrays
    wave = np.array(glb.wavelist, dtype='double')

    sl_path = os.environ['SL_CACHE_DIR']
    file_name = "%s/%04d/sl_pix_%06d.bin" % (sl_path, ipix_dir, ipix)
    if not os.path.exists(file_name):
        msg = 'Stray light data is not available for position (%s) on %s.' % (ra_dec_str, date_str)
        raise StraylightDataError(msg)
    nonzodi_bg = np.fromfile(file_name, dtype=nonzodi_pix_dtype, count=1)
    iday_pt = nonzodi_bg['iday_index'][0][iday-1]
    if iday_pt == -1:
        msg = 'Specified position ({}) is not observable on {}.\nUse of the ' \
              '<a href="https://jwst-docs.stsci.edu/jwst-other-tools/jwst-target-visibility-tools/jwst-general-target-visibility-tool-help" '\
              'target="_blank">General Target Visibility Tool (GTVT)</a> is recommended for determining the visibility'\
              ' of the specified position on the specified date'.format(ra_dec_str, date_str)
        raise StraylightPositionError(msg)
    zodi_sl_bgs = np.memmap(
        file_name,
        offset=nonzodi_pix_dtype.itemsize + zodi_sl_dtype.itemsize * iday_pt,
        mode='r',
        dtype=zodi_sl_dtype
    )
    stray_light_bg = np.array(zodi_sl_bgs['stray_light_bg'][0], dtype='double')
    return (wave, stray_light_bg)


def get_dateless_bg(ra, dec, level):
    """
    Provides equivalent straylight background spectra for pointing and day.

    Parameters
    ----------
    ra : double
        Right Ascension [degrees]
    dec : double
        declination [degrees]
    level : string
        Single character L,M,H

    Returns
    -------
    tuple of:
    doy : integer
        Day of year in 2020 where the backgrounds match the specified level.
        L,M,H = 10%, 50%, 90%.
    wave : numpy.ndarray
        Standard wavelengths of background values [microns]
    infield_bg : numpy.ndarray
        Infield background spectra from zodi, cib, and ism on Day of Year 2020 [MJy/str].
    stray_light_bg : numpy.ndarray
        Equivalent infield background sptectrum from stray light on Day of Year 2020 [MJy/str].
    thermal_wave : numpy.ndarray
        Wavelengths of equivilant thermal background values [microns].
    thermal_bg : numpy.ndarray
        Equivalent infield thermal background spectrum from JWST thermal self emission [MJy/str].
    """

    if level == 'L':
        ilevel = 0
    elif level == 'M':
        ilevel = 1
    elif level == 'H':
        ilevel = 2
    else:
        msg = 'Input level parameter, %s, is not L, M, or H.' % (level)
        raise DatelessBGError(msg)

    # Find the HEALPix number and the subdirectory
    ipix = ang2pix_ring(glb.nside, ra*glb.D2R, dec*glb.D2R)
    ipix_dir = ipix // 100

    # Set up the numpy arrays
    wave = np.array(glb.wavelist, dtype='double')

    sl_path = os.environ['SL_CACHE_DIR']
    file_name = "%s/dateless_background_index.bin" % (sl_path)
    if not os.path.exists(file_name):
        msg = 'dateless_background_index.bin file does not exist in %s.' % (sl_path)
        raise StraylightDataError(msg)
    dateless_index = np.memmap(
        file_name,
        offset=dateless_index_dtype.itemsize * ipix,
        mode='r',
        dtype=dateless_index_dtype
    )
    iday = dateless_index['dateless_index'][0][ilevel]
    file_name = "%s/%04d/sl_pix_%06d.bin" % (sl_path, ipix_dir, ipix)
    if not os.path.exists(file_name):
        msg = 'Dateless background data is not available for position (%s, %s).' % (ra, dec)
        raise DatelessBGError(msg)
    nonzodi_bg = np.fromfile(file_name, dtype=nonzodi_pix_dtype, count=1)
    iday_pt = nonzodi_bg['iday_index'][0][iday-1]
    # Get the spectra for the day
    zodi_sl_bgs = np.memmap(
        file_name,
        offset=nonzodi_pix_dtype.itemsize + zodi_sl_dtype.itemsize * iday_pt,
        mode='r',
        dtype=zodi_sl_dtype
    )
    nonzodi_bg = np.array(nonzodi_bg['nonzodi_bg'][0], dtype='double')
    zodi_bg = np.array(zodi_sl_bgs['zodi_bg'][0], dtype='double')
    infield_bg = nonzodi_bg + zodi_bg
    stray_light_bg = np.array(zodi_sl_bgs['stray_light_bg'][0], dtype='double')
    [thermal_wave, thermal_bg] = get_thermal_background()
    return (iday, wave, infield_bg, stray_light_bg, thermal_wave, thermal_bg)


def get_thermal_background():
    """
    Provides equivalent thermal background spectra. (renamed from get_thermal_bg to avoid confusion)

    Returns
    -------
    tuple of:
    wave : numpy.ndarray
        Wavelengths of background values [microns]
    thermal_bg : numpy.ndarray
        Equivalent in field background from observatory thermal emission.
    """

    sl_path = os.environ['SL_CACHE_DIR']
    file_name = "%s/%s" % (sl_path, THERMAL_FNAME)
    f = open(file_name, 'r')
    lines_all = f.readlines()
    f.close()
    lines = [l for l in lines_all if l[0] != '#']  # allow any number of comment lines
    sep = ',' if ',' in lines[0] else ' '         # allow space- or comma-separated files

    thermal_wave = []
    thermal_list = []

    for aline in lines:
        (awave, athermal) = aline.strip().split(sep)
        awave = float(awave)
        athermal = float(athermal)
        thermal_wave.append(awave)
        thermal_list.append(athermal)

    # Set up the numpy arrays
    wave = np.array(thermal_wave, dtype='double')
    thermal_bg = np.array(thermal_list, dtype='double')
    return (wave, thermal_bg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from importlib import metadata
import os

__version__ = metadata.version('pandeia.engine')

def pandeia_version():
    pandeia_data = "ENVIRONMENT VARIABLE UNSET."
    if "pandeia_refdata" in os.environ:
        refdata_path = os.environ['pandeia_refdata']
        try:
            with open("{}/VERSION_PSF".format(refdata_path)) as fp:
                # last char is a newline
                pandeia_data = fp.readline().strip()
        except OSError:
            pandeia_data = "INVALID INSTALLATION. CHECK PATH."

    cdbs_tree = "ENVIRONMENT VARIABLE UNSET."
    if "PYSYN_CDBS" in os.environ:
        cdbs_path = os.environ['PYSYN_CDBS']
        # CDBS Trees are not guaranteed to have versioning, so also check for
        # the existence of the directory.
        if os.path.isfile("{}/pyetc_cdbs_version".format(cdbs_path)):
            with open("{}/pyetc_cdbs_version".format(cdbs_path)) as fp:
                # the overall version is only the first line of this file
                cdbs_tree = fp.readline().strip()
        elif os.path.isdir(cdbs_path):
            cdbs_tree = cdbs_path
        else:
            cdbs_tree = "INVALID INSTALLATION. CHECK PATH."
    print("Pandeia Engine version:  {}".format(__version__))
    print("Pandeia RefData version:  {}".format(pandeia_data))
    print("Synphot Data:  {}".format(cdbs_tree))
    # If install is not run from a git repository, `git describe --tags` will 
    # return error code 128 which causes subprocess to raise a CalledProcessError
    # and the DEVBUILD file will not be updated by setup.py
    #Insurance if we forget to take this out
    if "dev" in __version__:
        # find the file
        buildfile = [p for p in metadata.files('pandeia.engine') if 'DEVBUILD' in str(p)][0]
        with open(buildfile.locate()) as fp:
            print("Development Build {}".format(fp.read()))



#! /usr/bin/env python
""" This module is the main script for BIT batch mode processing. This
is an edited version of the original copied from the pyetc codebase.

The plan is it takes an input file with one query per line and produces
an output file with one result per line.

It WILL BE structured to
facilitate parallel processing but for now can only be run serially.
"""
import io, optparse, os, pprint, sys, json
import signal
from multiprocessing import Pool
from datetime import datetime
import pysynphot

from pandeia.engine.helpers.bit.pyetc_util import log
from pandeia.engine.helpers.bit import translate
from pandeia.engine.helpers.bit import compute
from pandeia.engine.helpers.peng import peng3jeng
from pandeia.engine.perform_calculation import perform_calculation
from pandeia import engine

#from pyetc.etc_web.etc import versions

# ========================================================================

# process_query is the main parallelizable function.  It
# computes an ETC for one query at a time.  Since queries are
# independent, many can be run simultaneously.  Global references need
# to be handled specially for IPython's parallel processing: imports
# and cache.  Likewise exception handling is designed to trap all
# exceptions and funnel them back as either a response <message> or an
# exception string which is re-raised by the top-level loop, i.e. the
# serial controller.

RESULT_2_XML = {
    'brightest_pixel_rate': 'BrightestPixelRate',
    'detector_total_rate': 'ImageRate'
}

def make_element(name, value):
    return '<' + name + '>' + str(value) + '</' + name + '>'

def get_version_string():
    versions = '<VersionInfo>\n'
    versions += '    ' + make_element('Pandeia', engine.__version__) + '\n'
    versions += '    ' + make_element('Bit', engine.__version__) + '\n'
    versions += '    ' + make_element('Pysynphot', pysynphot.__version__) + '\n'
    versions += '    ' + make_element('Cdbs', os.environ['PYSYN_CDBS']) + '\n'
    versions += '    ' + make_element('DateProcessed', datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + '\n'
    versions += '</VersionInfo>'

    return versions

def create_xml_response(query, output_dict, detectors):
    # Grab the required values from the output dictionary.
    # TODO: This works for STIS, but for instruments with more than 1 detector we will also need to read 'b', 'c', etc
    bpr_name = 'brightest_pixel_rate'
    ir_name = 'detector_total_rate'
    bpr_xml_name = RESULT_2_XML[bpr_name]
    ir_xml_name = RESULT_2_XML[ir_name]

    found_detector = False
    for detector in detectors:
        if detector in output_dict['bop'][bpr_name]:
            bpr = output_dict['bop'][bpr_name][detector][0]
            ir = output_dict['bop'][ir_name][detector]
            found_detector = True

    if not found_detector:
        raise Exception("Calculation includes detector not found in detectors list.  May need to update the detectors list in bit.py")

    #bpr_wavelength = output_dict['bop'][bpr_name]['a'][1]
    #ir_wavelength = output_dict['bop'][ir_name]['a'][1]
    vr = "true"

    xml = '<Response>'
    xml += make_element(bpr_xml_name, bpr)
    xml += make_element(ir_xml_name, ir)
    xml += make_element('validRequest', vr)

    # Escape the ampersands and strip the new line character from the end of the query.
    #  The new line was causing the failure in ingestion.
    xml += make_element('query', query.replace('&', '&amp;').strip())
    xml += '</Response>'

    return xml

def get_detectors_list():
    data_dir = os.environ['pandeia_refdata']
    detectors = []

    # Open the HST telescope configuration file to read all of the instruments.
    with open(data_dir + '/hst/telescope/config.json') as tele_file:
        tele_dict = json.load(tele_file)

        # Loop through the instruments, and get all of the detectors.
        for instrument in tele_dict['instruments']:
            with open(data_dir + '/hst/%s/config.json' % instrument) as inst_file:
                inst_dict = json.load(inst_file)
                detectors.extend(inst_dict['detectors'])

    return detectors

def _process_query(query_str):
    """process_query() translates a query_str into an ETC,  runs it,
    and returns an XML response string containing requested variables.
    This is the main function designed for parallel processing,  hence
    it uses no global variables,  not even modules.
    """
    # Don't catch <control>-C in child processes
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    # First translate and run the ETC, returning exceptions as
    # <message> text.
    try:
        web_inputs, requests = translate.process_query(query_str, info=False)
        full_results = compute.compute_bright_object(web_inputs)
        message = None
    except Exception as excp:
        full_results = None
        requests = []
        message = repr(excp)

    # Next extract the requested variables from the results and format
    # XML.  Note that any exception in this phase prevents the
    # creation of the XML and <message> element used to report
    # exceptions above.  So these are actually Response-less XML-less
    # results.  Generally there should be no exceptions here since
    # even undefined request variables are already trapped and
    # handled.
    try:
        requested_xml = compute.get_partial_results(
            query_str, requests, full_results, message)
        return True, query_str, requested_xml
    except Exception as excp:   # Didn't get XML
        return False, query_str, str(excp)

def process_query(query_str):
    """Capture subprocess output to prevent scrambling during
    multiprocessing.
    """
    output = io.StringIO()
    sys.stdout, sys.stderr = output, output
    result = _process_query(query_str)
    sys.stdout, sys.stderr = sys.__stdout__, sys.__stderr__
    return result, output.getvalue().strip()

# ========================================================================

def version_string(subsystem):
    """Return a string identifying the version of `subsystem` being used."""
    if subsystem == "cdbs":
        return os.environ["PYSYN_CDBS"]
    else:
        retval = ''
        vers = versions.get_module_versions()
        if subsystem == 'bit':
            subsystem = 'pyetc'
        # get basic VCS revision
        retval = vers[subsystem]["rev"]
        # handle pyetc.* subsystems which no longer have their own version
        if subsystem.startswith('pyetc.') and retval == '':
            retval = vers['pyetc']["rev"]
        # handle non-pyetc subsystems which only have ver info in the 'str' field
        if retval == '' and vers[subsystem]['str'] not in ('', 'unknown'):
            return vers[subsystem]['str']
        # prepend version name if any
        if subsystem.startswith('pyetc') and vers[subsystem]['file'] not in ('', 'unknown'):
            vername = [pth for pth in vers[subsystem]['file'].split('/') if pth.startswith('pyetc_')][0]
            retval = vername+' '+retval
        # finally
        return retval

def get_and_log_version(subsystem):
    """Get a software version, return it, and report it in the log."""
    vers = version_string(subsystem)
    log.info("Version", subsystem, vers)
    return vers

def process_queries(inname, outname, n_compute_nodes=1):
    """process_queries() defines the top level loop over all the queries
    found in file `inname`,  writing XML results to file `outname`.
    When n_compute_nodes > 1,  process_queries is designed to parallel
    process small numbers of queries via the `mapper` function.
    """
    log.info("Processing with", n_compute_nodes, "compute nodes.")

    engine_vers = get_and_log_version("pyetc.engine")
    instruments_vers = get_and_log_version("pyetc.instruments")
    web_vers = get_and_log_version("pyetc.etc_web")
    bit_vers = get_and_log_version("bit")
    pysynphot_vers =get_and_log_version("pysynphot")
    cdbs_vers = get_and_log_version("cdbs")

    pool = Pool(n_compute_nodes)

    simple_file = open(outname,"w+")
    simple_file.write("<Responses>\n")

    simple_file.write("<VersionInfo>\n")
    simple_file.write("    <PyetcEngine>" + engine_vers + "</PyetcEngine>\n")
    simple_file.write("    <PyetcInstruments>" + instruments_vers + "</PyetcInstruments>\n")
    simple_file.write("    <PyetcWeb>" + web_vers + "</PyetcWeb>\n")
    simple_file.write("    <Bit>" + bit_vers + "</Bit>\n")
    simple_file.write("    <Pysynphot>" + pysynphot_vers + "</Pysynphot>\n")
    simple_file.write("    <Cdbs>" + cdbs_vers + "</Cdbs>\n")
    simple_file.write("    <DateProcessed>" +
        str(datetime.datetime.now()).split(".")[0] + "</DateProcessed>\n")
    simple_file.write("</VersionInfo>\n")

    # use 1 as a sentinel to tell us not to use any multiprocessing code
    parallel_process = n_compute_nodes > 1

    # chunks() returns n_compute_nodes-length lists generated by generator
    # processing by chunks helps generate incremental output.
    query_generator = translate.queries(inname)
    if not parallel_process:
        for query in query_generator:
            log.info("BIT Query:", repr(query))
            (xmlok, query, requested_xml), output = process_query(query)
            log.write(output, eol="")
            if xmlok:
                simple_file.write(requested_xml)
            if not xmlok or "<message>" in requested_xml:
                log.error("Exception:", requested_xml)
    else: # parallel_process
        for chunk in chunks(query_generator, n_compute_nodes):
            # mapper() applies process_query to each element of chunk
            # this is where parallel processing occurs.
            # "chunk" is a list of n_compute_nodes queries
            try:
                for (xmlok, query, requested_xml), output in pool.map(process_query, chunk):
                    log.info("BIT Query:", repr(query))
                    log.write(output, eol="")
                    if xmlok:
                        simple_file.write(requested_xml)
                    if not xmlok or "<message>" in requested_xml:
                        log.error("Exception:", requested_xml)
            except KeyboardInterrupt as e:
                pool.terminate()
                raise

    simple_file.write("</Responses>\n")

def chunks(generator, chunk_len):
    """Chop generator of large sequence into smaller list chunks since
    IPython parallel processing doesn't work on generators, just
    things with __len__().  Also yield multiple smaller sequences to
    avoid pre-computing 70k queries.
    """
    chunk = []
    for item in generator:
        chunk.append(item)
        if len(chunk) >= chunk_len:
            yield chunk
            chunk = []
    yield chunk

# ========================================================================

def query_to_peng(query_str):
    """ Translate a single query string to pyetc engine inputs (peng) dict. """
    web_inputs, requests = translate.process_query(query_str, info=False)
    engine_inputs = compute.web_to_engine_inputs(web_inputs)
    return web_inputs, engine_inputs

def debug_queries(query_list):
    """ Loop over them. """
    for q in query_list:
        print('\n'+q)
        debug_query(q)
        junk = input('Okay?')

def run_query(query_str, detectors, only_response=False, debug=False):
    """Process and dump a single query or a file of queries.

    Parameters
    ----------
    query_str: string
        The query to process.
    detectors: list
        The list of HST detectors ('ccd', 'nuvmama', etc.)
    only_response: bool
        If False, this function will return a complete XML BIT table (including versions) for the query.
        If True, this function will only return the <Response>...</Response> element for the query.

    Returns
    -------
    xml_response: string
        The XML response for the query, formatted as specified by the only_response parameter.
    """
    log.VERBOSE_FLAG = True

    web_inputs, engine_inputs = query_to_peng(query_str)

    # HERE we call peng3jeng
    jeng_dict = peng3jeng.pyetc_in_dict_to_pandeia_in_dict(engine_inputs, None, pweb_dict=web_inputs)

    if debug:
        log.verbose("\npandeia engine inputs:\n", pprint.pformat(jeng_dict), "\n")

    # Run the Pandeia calculation on the jeng input.
    jeng_results = perform_calculation(jeng_dict, webapp=True, dict_report=True)

    if debug:
        log.verbose("\npandeia engine outputs:\n", pprint.pformat(jeng_results), "\n")

    # Get the versions string.
    versions_str = get_version_string()

    # Extract what we need from the results and convert to XML.
    xml_response = ''
    if not only_response:
        xml_response = '<Responses>\n'
        xml_response += versions_str + '\n'

    xml_response += create_xml_response(query_str, jeng_results, detectors) + '\n'

    if not only_response:
        xml_response += '</Responses>'

    if debug:
        log.info("XML:", xml_response)

    return xml_response

# ========================================================================


def main():
    """Process a file of bit queries, `input`,  into a file of XML results,
    `output`.
    """
    parser = optparse.OptionParser(usage="usage: %prog [options] <input> <output>")

#   parser.add_option(
#       "-n", "--parallel-process-nodes", dest="parallel_compute_nodes",
#       help="Specify CPU count for parallel processing,  default=1.",
#       metavar="NODES", default=1)

    parser.add_option(
        "-q", "--query", dest="query",
        help="URL query string for a BIT request",
        metavar="QUERY", default=None)

    parser.add_option("-f", "--queryfile", dest="query_file", help="Input file with one query per line", default=None)

    parser.add_option("-o", "--outdirfn", dest="output_dir_filename", help="Full path and filename of output table", default="./bit_table.xml")

    parser.add_option('-D', '--direct', dest='direct', action='store_true',
                      help="Direct call - do not employ pdb wrapper")

#   options, args = log.handle_standard_options(parser, sys.argv, show_info=False)
    options, args = parser.parse_args(sys.argv)

    # Get the list of detectors.
    detectors = get_detectors_list()

    if options.query_file is not None:
        # A query file is set, so read the queries from there.
        query_file = open(options.query_file, 'r')
        queries = query_file.readlines()
        query_file.close()

        xml_responses = '<Responses>\n'
        xml_responses += get_version_string()

        # Process each of the queries.
        for query in queries:
            try:
                response = run_query(query, detectors, only_response=True)
                xml_responses += response
            except Exception as e:
                log.info(e)

        xml_responses += '</Responses>'

        out_file = open(options.output_dir_filename, 'w')
        out_file.write(xml_responses)
        out_file.close()
    elif 0 and options.query is None:
        # cds: TRANSLATE THIS to get rid of standard_run()
        log.standard_run("process_queries(args[1], args[2], "
                         "int(options.parallel_compute_nodes))",
                         options, globals(), locals())
    else:
        if 1 or options.direct: # run ONLY direct for a hot minute to get this started
            try:
                response = run_query(options.query, detectors)
                out_file = open(options.output_dir_filename, 'w')
                out_file.write(response)
                out_file.close()
            except Exception as e:
                log.info(e)
        else:
            log.standard_run("run_query(options.query)",
                             options, globals(), locals())
    log.write()
#   log.standard_status()

#
# For example, to run a single query (to debug say), it is
# % ./bit.py -D -q 'type=bot&amp;magnitude=10&amp;config=.....'
#
# Or, to generate a table from a file containing multiple queries, use:
#  python bit.py -f /path/to/query_file.txt -o /path/to/output_directory/output_table.xml
#  bit_table.xml will be written out to the specified directory.
if __name__ == "__main__":
    main()



###
# not part of the interface!  used by subsidiary instruments
default_mode_names = {
    'imaging'           : 'Imaging',
    'rampfilter'        : 'Ramp Filter',
    'spectroscopicacq'  : 'Spectroscopy Target Acquisition',
    'spectroscopic'     : 'Spectroscopy',                       # BUG: why different?
    'targetacquisition' : 'Imaging Target Acquisition',
}

default_short_mode_names =  {
    'imaging'           : 'im',
    'rampfilter'        : 'rf',
    'spectroscopicacq'  : 'sa',
    'spectroscopy'      : 'sp',
    'targetacquisition' : 'ta',
}




#TODO web cleanup:
# there is some duplication in between WFC3 UVIS and IR modules.

import pandeia.engine.helpers.bit.instruments
import pandeia.engine.helpers.bit.pyetc_util as util

#
# read a dictionary from one of the .dat files
def get_data( name ) :
    return pandeia.engine.helpers.bit.instruments.get_instrument_data( 'wfc3uvis', name )

FILTERS = get_data("config")["filters"]


FILTER_MAP = {
    'narrow': 'wfc3_filter_n',
    'medium': 'wfc3_filter_m',
    'wide':   'wfc3_filter_w',
    'quad':   'wfc3_filter_q',
}

def instrument_from_form(science_mode, form_input):
    """Function to extract values of interest from the web form
    data and return a smaller dictionary which will be used to
    update the engine_input dictionary.

    """

    idict = dict( detector=form_input['detector'].lower(),
                  nreads = form_input['crsplit'],
                  )
    #check that crsplit is valid
    if idict['nreads'] < 1:
        raise ValueError("CRSPLIT must be greater than 0")

    #Web output dictionary that will be stuffed into db
    idict['wodict'] = wodict = util.stable_dict()
    #all wfc3uvis detectors are ccd-like
    wodict['detector_type'] = 'ccd'

    # Only show post flash in certain situations, see #922
    wodict["show_post_flash"] = False

    if science_mode == 'imaging' or science_mode == 'scimaging':
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'], myfilter = make_wfc3uvis_imaging_obsmode(form_input)
        wodict['filter_info']=FILTERS[myfilter]
        wodict['filter_name']=myfilter
        idict['filter'] = myfilter.lower()
        idict['post_flash_electrons'] = form_input['post_flash_wfc3']
        wodict["show_post_flash"] = True

        return idict

    if science_mode == 'spectroscopic':
        #make an obsmode string that can be submitted to pysynphot
        idict['obsmode'] = make_wfc3uvis_spectroscopic_obsmode(form_input)
        idict['obswave']=form_input['obswave']
        idict['filter']='g280'
        idict['grating']='g280'
        return idict
    else:
        raise ValueError('%s mode not implemented for wfc3uvis'%science_mode)
    return idict


def make_wfc3uvis_imaging_obsmode(form_input):

    # figure out what kind of filter was selected
    ftype = form_input['wfc3_filter_type']
    myfilter = form_input[FILTER_MAP[ftype]]
    obsmode='wfc3,'+form_input['detector'].lower() + ','  + myfilter
    return obsmode, myfilter

def make_wfc3uvis_spectroscopic_obsmode(form_input):
    return  'wfc3,'+form_input['detector'].lower() + ',g280'



# COS limits.

# This file is read when the instruments/data.py file is executed
# as part of the instrument module import. The resulting dictionary
# is used to initialize instances of Limit that will be used to
# check any subsequent COS calculation.

# allsub defines what the magic value "all" should expand into
allsub = {'mode':['imaging', 'spectroscopic', 'targetacquisition', 'spectroscopicacq'],
          'detector' : ['nuv', 'fuv']
         }

# limit_spec contains the actual limit definitions. It is a
# dictionary {limit shortname : {definition of limit} }.
# See user's guide for detailed requirements of the limit definition.

# Message strings may use the following variable names for substitution:
#  varname = value of the variable specified in "varname"
#  threshold = value specified in "threshold"
#  percentval = percentage expression of (1-subthresh)
#
limit_spec = {
    'maxtime' : {
       'varname' : 'time',
       'mode'    : 'targetacquisition',
       'detector': 'nuv',
       'threshold': 1500,
       'message': '''The imaging target acquisition procedure requires that two
                     images be obtained and this should occur within one orbital
                     visibility period (typically about 50 min or 3000 sec). The
                     time required for your target (%(varname)g) exceeds %(threshold)g sec''',
    }, #end maxtime

    'minbuftime' : {
       'varname' : 'buffer_time',
       'mincheck': True, #NOTE, subthresh is weird for minchecks. Need tests & careful logic. Or just break it out for simplicity?
       'mode'    : ['spectroscopic','imaging'],
       'detector': 'all',
       'subthresh' : 1.3875, #whatever leads to 111,
       'submessage' : '''Buffer fill time %(varname)g is less than %(subthresh_value)g seconds.
                         Data loss may result if observation is performed in TTAG mode.''',
       'threshold' : 80,
       'varieties' : (
         {
          'mode' : 'spectroscopic',
          'message': 'Buffer fill time %(varname)g is less than minimum %(threshold)g seconds.',},
         {'mode' : 'imaging',
          'message' : '''Because buffer fill time is less than %(threshold)g sec, this observation can
                         only be performed in ACCUM mode.''',
          }
          )


    }, #end minbuftime


### Source for following brightpix, brightsegm and brightimg limits:
### COS Instrument Handbook 2.0 January 2010 for Cycle 18
### Table 9.1, "COS Count Rate Screening Limits",
### section 9.2, p 90.


    'brightpix' : {
       'varname' : 'brightest_pixel_rate',
       'message' : 'Total count rate per pixel %(varname)g exceeds bright limit %(threshold)g.',

       'varieties' : (
           { #fuv: spec limits only
             #Updated to support future cenwave dependence: PR 72344
            'detector': 'fuv',
            'mode'    : ['spectroscopic', 'spectroscopicacq'],
             #Updated in accordance with new EE tables, per PR 72859
            'threshold': ('cenwave',
                          {1055: 0.2,
                           1096: 0.2,
                           'default' : 0.666667,},
                           ),
            },

           { #nuv: spec limits
            'detector': 'nuv',
            'mode'    : ['spectroscopic','spectroscopicacq'],
            'threshold': 70,},
           { #nuv: imaging limits
             'detector': 'nuv',
             'mode'    : ['imaging','targetacquisition'],
             'threshold': 50,},
           ),
    }, #end brightpix


    'brightseg' : {
       'varname' : 'brightest_segment_rate',
       'mode'    : ['spectroscopic','spectroscopicacq'],
       'varieties' : (
           { #fuv
             'detector': 'fuv',
             'threshold': 15000,},
           { #nuv
              'detector': 'nuv',
              'threshold': 30000,},
           ),
       'message' : '''Segment countrate %(varname)s exceeds segment/stripe global count rate
                      limit of %(threshold)g counts per second for non-variable sources.''',
       'subthresh': 0.4,
       'submessage':'''Segment countrate %(varname)s exceeds segment/stripe limit for irregularly-variable
                       sources. (The segment limit for irregularly-variable sources is %(percentval)g percent
                       of global count rate limit, i.e., %(percentval)g percent of %(threshold)g counts per
                       second.)''',
    }, #end brightseg

    'brightimg' : {
       'varname' : 'detector_total_rate',
       'mode': ['imaging','targetacquisition'],
       'detector': 'nuv',
       'threshold': 170000,
       'message': '''Observation countrate %(varname)g exceeds NUV imaging global screening
                     count rate limit of %(threshold)g counts per second for non-variable sources.''',
       'subthresh': 0.4,
       'submessage': '''Observation countrate %(varname)g exceeds the countrate limit for
                        irregularly variable sources. (The countrate limit for irregularly
                        variable sources is %(percentval)g percent of NUV imaging global
                        screening count rate limit, i.e., %(percentval)g percent of %(threshold)g
                        counts per second.)'''
    }, #end brightimg

    'maxleak' : {
        'varname' : 'total_leak',
        'message' : '''This observation has a significant filter leak where
                       %(varname).3g percent of the signal comes from outside the prescribed
                       filter bandpass limits.''',
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 5.0,
    }, # end maxleak

    'zerotargetflux' : {
        'varname' : 'target_total_rate',
        'mincheck' : True,
        'mode' : 'imaging',
        'detector' : 'all',
        'per_read' : False,
        'threshold' : 1.E-25,
        'varieties': (
            {'mode'   : 'imaging',
             'message': '''Zero target flux in the observation bandpass''',},
            {'mode'   : 'spectroscopic',
             'message': '''Zero target flux at the observation wavelength''',},
        ),
       }, # end zerotargetflux

    # this is not a real limit, but a warning associated with the detector/science mode
    # combination. We fake it as a zero-threshold limit so the existing infrastructure
    # for limit processing can be re-used. This infrastructure creates as main product
    # a warning message with the desired properties.
    'extractbox' : {
       'varname' : 'target_total_rate',
       'mode'    : ['spectroscopic'],
       'varieties' : (
           {
             'detector': 'nuv',
             'threshold': 0,},
           ),
       'message' : '''Note that for faint sources, the S/N calculated by the ETC for 
       COS/NUV observations may only be achieved by using a smaller spectral extraction 
       height than the one used in the standard reduction.  Please contact the STScI 
       Help Desk for more information.  Furthermore, G285M calculations severely 
       underestimate the exposure times required.   Please see section 2.4 of the COS 
       Instrument Handbook for details.''', 
    }, #end extractbox

} # end limit_spec



'''Map the extraction form parameters into a form that the etc_engine can use.
'''

import os

#from ...main import util, tools
from pandeia.engine.helpers.bit.pyetc_util import read_dict
import pandeia.engine.helpers.bit.instruments as instruments


mapfile = os.path.join(os.path.dirname(__file__), 'etc_web_data/extraction.dat')
EXTRACTION_MAP = read_dict(mapfile)


def _handle_scan_mode(ans, fd, shape):

    # Here, we handle the 'width' value.
    #
    # 'varname' is the name of the HTML field that contains the relevant extraction parameter:
    # - width for imaging. The height is computed elsewhere.
    #   - height (scan length) for spectroscopy. The width is implicitly assumed to be 1 resel.
    #
    # The logic here assumes that the scan direction is parallel to the "height" direction.
    # We have to handle the fact that the extraction rectangle dimensions in spectroscopy are
    # reversed wrt imaging (an engine API hurdle).
    if fd['science_mode'] == 'scimaging':
        # in imaging, the dimension that matters for PSF-extraction purposes is the width (perpendicular
        # to the scan direction). It is used to index and interpolate the EE table. It is directly read
        # from the page. It will end up generating a rectangle spec such as [5.0, 82.57]. In here, the
        # first element (5.0) is the width read from the page, and the second element (82.57) is computed
        # elsewhere.
        varname = "extractionRegionScWidth"

    else:
        # in spectroscopy, there is no dimension associated with EE table weighting. The width is implicitly
        # 1 resel, and it is inserted into the extraction region specification later on. The only relevant size
        # is the scan length itself, which is the height of the extraction region. The same scan length used
        # in the above comment would then generate a rectangle spec like [82.57, 1.0], meaning 82.57 pixels
        # high and 1 resel wide.
        varname = "extractionRegionScHeight"

    # Look up the size (in pixels) from the variable name.
    size = fd[varname]
    ans['size'] = size

    # Scan mode adds further complexity to rectangles. Here, we handle the 'height' value.
    if 'rectangle' in shape:

        height = fd['extractionRegionScHeight']

        # 'extractionRegionScHeight' is a drop-down selector
        # that has entries of 'scanLength' (assumed to be in
        # arcsec) and '1' (assumed to be in pixels).

        # Only run the clause in case the drop-down was selected
        # as 'scanLength'. The '1' case is handled trivially by
        # just holding the value, unaltered, in the 'height' variable.
        if height == 'scanLength':

            # scan length can be gotten from either of two page variables
            # Which one to use is controlled by the 'simmode' radio button.
            # This radio button can have the values of either 'SNR', or
            # 'SNR_from_rate'. We *have* to use 'SNR' for the button case
            # that allows the user to enter the Time value, so as to re-use
            # the existing web code logic for stare mode.

            # 'SNR_from_rate' means: "compute SNR given a Time value derived
            # from a scan rate and a scan length".
            if fd['simmode'] == 'SNR_from_rate':
                height = fd['scLengthFromRate']
            # 'SNR' means what it always meant in stare mode: "compute SNR
            # given a Time derived from a page field"
            if fd['simmode'] == 'SNR':
                height = fd['scLengthFromTime']

        # need plate scale to convert extraction height to pixels.
        instrument_name = fd['instrument']
        instconfig = tools.find_instrument('hst', instrument_name)
        dconfig = instconfig.data.get_dconfig(fd['detector'])
        pixel_height = dconfig['pixel_height']

        # in scan mode, when height=1, it really means height=1 pixel.
        # Any other value, it's assumed height is in arcsec and thus
        # must be converted to pixels before feeding to the engine.
        height = float(height)
        if height != 1.0:
            height /= pixel_height

        # Extraction box in imaging is reversed wrt spectroscopy...
        if fd['science_mode'] == 'scimaging':
            ans['size'] = [float(size), float(height)]
        else:
            ans['size'] = [float(height), 1.]


#TODO web cleanup:
# this function is too long; should be broken into smaller pieces. Obvious candidate is the scan-mode-specific code.

def extraction_info(formdata, scan_mode):
    """ Map form data into extraction data, with instrument-specific
    lookups to provide 'default' (aka recommended) values.
    """
    fd = formdata
    ans = {}

    # Copy in the gyromode (for now at least)
    ans['gyromode'] = fd['gyromode']

    # Look up the source type (point or extended)
    target_type = fd['fsourceType']

    # Based on that, look up the source diameter
    # (in arcsec) and the aperture shape.
    shape = _handle_target_type(ans, fd, target_type)

    #TODO web cleanup:
    # probably need to add a hook for instrument-specific customizations here

    # BUG: in the meantime there is a hardcoded hack for
    # COS specacq mode which needs an extra parameter
    if (fd['instrument'] == 'cos' and
        fd['science_mode'] == 'spectroscopicacq'):
        # then we need an extra variable
        # (only for NUV, because it makes no difference for FUV)
        ans['acq_mode'] = fd.get('nuvMode','')
        ans['stripe'] = fd.get('Stripe','')

    # (Source type, aperture shape) key the extraction_map, which
    # determines which variable to look in. Disperser is used in
    # a subset of cases.
    disperser_ = fd.get('disperser', '')

    if shape == 'default':
        # Then look up the instrument-specific defaults, which
        # depend on source type and mode as well as detector.
        ans['shape'],ans['size'] = _get_inst_extraction_defaults(fd['instrument'],
                                                                 fd['science_mode'],
                                                                 fd['detector'],
                                                                 disperser_,
                                                                 target_type)
    else:
        # Otherwise, read it from the standard map
        ans['shape'], varname = EXTRACTION_MAP[(target_type, shape)]

        # there might be one level of indirection. This happens for
        # now only with wfc3 uvis and ir spectroscopic mode.
        if varname == 'xRegionType':
            rt = fd[varname]
            if rt == 'default':
                ans['shape'],ans['size'] = _get_inst_extraction_defaults(fd['instrument'],
                                                                         fd['science_mode'],
                                                                         fd['detector'],
                                                                         disperser_,
                                                                         target_type)
                # invalidate so it won't be used after here.
                varname = ''
            else:
                ans['shape'], varname = EXTRACTION_MAP[(target_type, rt.lower())]

        # Look up the size (in pixels) from the variable name. This will work for
        # most variable names, but fails with the special variable names used in
        # scan mode. Scan mode needs special handling.
        if varname in fd:
            size = fd[varname]
            ans['size'] = size

            # spectroscopic pages deliver the extraction aperture size as a
            # comma-separated string. Convert it to the engine API format, a
            # tuple. Conversions look unsafe but will be done only on values
            # provided by the html code itself. In case one envisions other
            # uses for this code, a safer version must be put in place.
            if type(size) == type(""):
                size = size.split(",")
                ans['size'] = (float(size[0]), float(size[1]))

        # Scan mode adds complexity to input page fields.
        if scan_mode:
            _handle_scan_mode(ans, fd, shape)

    #TODO web cleanup:
    # an instrument hook here would be handy.

    if (fd['instrument'] == 'stis' and
        fd['science_mode'] == 'targetacquisition' and
        "peak" in fd['ccdMode'].lower()):

        ans['shape'] = 'square'
        ans['size']  = 32

    # #815: IMAGE / SEARCH modes for COS image target acquisition.
    if (fd['instrument'] == 'cos' and
        fd['science_mode'] == 'targetacquisition'):

        if fd.get('cosAcqMode','') == 'search':
            ans['shape'] = 'rectangle'
            ans['size']  = (220,470)

    #That's it.
    return ans


def _handle_target_type(ans, fd, target_type):

    ans['stype'] = target_type

    if target_type == 'point':
        ans['sdiameter'] = 0.
        shape = fd.get('xRegionType', '').lower()

    elif target_type == 'extended':
        ans['sdiameter'] = fd['fdiameter']  # fdiameter is in arcsec
        shape = fd.get('xRegionExtendedType', '').lower()

    else:
        raise ValueError('Unknown type of extraction region')

    return shape


def _get_inst_extraction_defaults(instrument,
                                  mode,
                                  detector,
                                  disperser,
                                  target_type): #point/extended



    """ Return info from instrument-specific dictionary.
    Be smart and cache results so we don't have to repeatedly read the same
    dictionary file.

    """
    # TODO:
    #It's possible this should call out to a module in the specific
    #instrument, because the default dictionaries are still being shoehorned
    #into a standard.

    # Import the correct module
    mod = instruments.import_instrument_module(instrument,
                                                   '_web')
    # (Efficiently) get the data we need
    EXTRACTION_DEFAULTS = mod.get_data('extraction_defaults')

    # Now it has the right dictionary loaded: look up the
    # default values.
    # Some instruments may have the default extraction regions
    # keyed with 3 parameters, most are keyed with 2.
    try:
        defval = EXTRACTION_DEFAULTS[(target_type, mode)]
    except KeyError:
        index = disperser.find('_0')
        grating = disperser[index+2:]
        defval = EXTRACTION_DEFAULTS[(target_type, mode, grating)]

    # Some instruments have a further level of nesting by dict;
    # others are simpler.
    if isinstance(defval, dict):
        shape,size = defval[detector]
    else:
        shape, size = defval

    return shape, size
